from syngen.ml.validation_schema.validation_schema import (  # noqa: F401;
    TrainingSettingsSchema,
    InferSettingsSchema,
    CSVFormatSettingsSchema,
    ExcelFormatSettingsSchema,
    GlobalSettingsSchema,
    ConfigurationSchema,
    KeysSchema,
    ValidationSchema,
    SUPPORTED_EXCEL_EXTENSIONS,
    ReportTypes
)



from typing import Union, List, Optional, Dict, Literal
from abc import ABC
from itertools import combinations
from collections import Counter
import re
import random

import tqdm
from sklearn.cluster import KMeans
from sklearn.metrics import davies_bouldin_score

from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier
from sklearn.ensemble import RandomForestRegressor
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler, StandardScaler
from sklearn.metrics import r2_score, accuracy_score
import matplotlib
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.patches import Patch
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.stats as st
from scipy.stats import kurtosis, skew
import seaborn as sns
from slugify import slugify
from loguru import logger

from syngen.ml.utils import (
    timestamp_to_datetime,
    fetch_config
)
from syngen.ml.vae.models.features import KURTOSIS_THRESHOLD

matplotlib.use("Agg")


class BaseMetric(ABC):
    def __init__(
        self,
        original: pd.DataFrame,
        synthetic: pd.DataFrame,
        plot: bool,
        reports_path: str,
    ):
        self.original = original
        self.synthetic = synthetic
        self.reports_path = reports_path
        self.plot = plot
        self.value = None

    def get_value(self) -> Union[float, List[float]]:
        return self.value


class JensenShannonDistance(BaseMetric):
    def __init__(
        self,
        original: pd.DataFrame,
        synthetic: pd.DataFrame,
        plot: bool,
        reports_path: str,
    ):
        super().__init__(original, synthetic, plot, reports_path)
        self.cmap = LinearSegmentedColormap.from_list(
            "rg", ["#96195C", "#C13666", "#B24E89", "#9075C1", "#3F93E1", "#E8F4FF"]
        )

    def calculate_all(self, categorical_columns: List[str]):
        self.original[categorical_columns] = self.original[categorical_columns].fillna("")
        self.synthetic[categorical_columns] = self.synthetic[categorical_columns].fillna("")
        self.heatmap, self.labels = self.__compute_vs_columns(categorical_columns)

        if self.plot:
            plt.clf()
            sns.set(
                rc={
                    "xtick.labelsize": self.heatmap.shape[0],
                    "ytick.labelsize": self.heatmap.shape[0],
                    "ytick.major.size": 1.5,
                    "ytick.major.width": 0.2,
                },
                font_scale=2,
            )
            heatmap = sns.heatmap(
                self.heatmap,
                xticklabels=self.labels,
                yticklabels=self.labels,
                cmap=self.cmap,
                vmin=0.0,
                vmax=1.0,
                center=0.5,
                annot=False,
            )
            heatmap.figure.set_size_inches(self.heatmap.shape)
            heatmap.axes.tick_params(axis="x", rotation=90)
            heatmap.axes.tick_params(axis="y", rotation=0)
            heatmap.figure.tight_layout()
            plt.savefig(
                f"{self.reports_path}/accuracy_heatmap.svg",
                bbox_inches="tight",
                format="svg",
            )
            plt.clf()

    @staticmethod
    def calculate_heatmap_median(heatmap):
        heatmap_no_diag = heatmap[~np.eye(heatmap.shape[0], dtype=bool)].reshape(
            heatmap.shape[0], -1
        )
        heatmap_median = np.nanmedian(heatmap_no_diag)  # ignores nan when calculating median
        return heatmap_median

    def _calculate_pair_continuous_vs_continuous(self, first_column, second_column):
        original_score = self.__jensen_shannon_distance(
            self.original[first_column].fillna(self.original[first_column].mean()),
            self.original[second_column].fillna(self.original[second_column].mean()),
        )

        synthetic_score = self.__jensen_shannon_distance(
            self.synthetic[first_column].fillna(self.synthetic[first_column].mean()),
            self.synthetic[second_column].fillna(self.synthetic[second_column].mean()),
        )
        return 1 - abs(original_score - synthetic_score)

    def _calculate_pair_categ_vs_continuous(self, first_column, second_column):
        map_dict = {
            k: i + 1
            for i, k in enumerate(
                set(self.original[first_column]) | set(self.synthetic[first_column])
            )
        }
        original_score = self.__jensen_shannon_distance(
            self.original[first_column].map(map_dict),
            self.original[second_column].fillna(self.original[second_column].mean()),
        )

        synthetic_score = self.__jensen_shannon_distance(
            self.synthetic[first_column].map(map_dict),
            self.synthetic[second_column].fillna(self.synthetic[second_column].mean()),
        )
        return 1 - abs(original_score - synthetic_score)

    def _calculate_pair_categ_vs_categ(self, first_column, second_column):
        map_dict_first = {
            k: i + 1
            for i, k in enumerate(
                set(self.original[first_column]) | set(self.synthetic[first_column])
            )
        }
        map_dict_second = {
            k: i + 1
            for i, k in enumerate(
                set(self.original[second_column]) | set(self.synthetic[second_column])
            )
        }

        original_score = self.__jensen_shannon_distance(
            self.original[first_column].map(map_dict_first),
            self.original[second_column].map(map_dict_second),
        )

        synthetic_score = self.__jensen_shannon_distance(
            self.synthetic[first_column].map(map_dict_first),
            self.synthetic[second_column].map(map_dict_second),
        )
        return 1 - abs(original_score - synthetic_score)

    def _calculate_pair_continuous_vs_categ(self, first_column, second_column):
        map_dict = {
            k: i + 1
            for i, k in enumerate(
                set(self.original[second_column]) | set(self.synthetic[second_column])
            )
        }

        original_score = self.__jensen_shannon_distance(
            self.original[first_column].fillna(self.original[first_column].mean()),
            self.original[second_column].map(map_dict),
        )

        synthetic_score = self.__jensen_shannon_distance(
            self.synthetic[first_column].fillna(self.synthetic[first_column].mean()),
            self.synthetic[second_column].map(map_dict),
        )
        return 1 - abs(original_score - synthetic_score)

    def __compute_vs_columns(self, categorical_columns: List[str]):
        valid_cols = [col for col in self.original.columns if "word_count" not in col]
        mask = list(map(lambda col: col in categorical_columns, valid_cols))
        heatmap_matrix = []
        for i, c in enumerate(valid_cols):
            row = []
            for j, c_ in enumerate(valid_cols):
                if mask[i]:
                    if mask[j]:
                        acc = self._calculate_pair_categ_vs_categ(c, c_)
                    else:
                        acc = self._calculate_pair_categ_vs_continuous(c, c_)
                else:
                    if mask[j]:
                        acc = self._calculate_pair_continuous_vs_categ(c, c_)
                    else:
                        acc = self._calculate_pair_continuous_vs_continuous(c, c_)

                row.append(acc)
            heatmap_matrix.append(row)

        return np.array(heatmap_matrix), valid_cols

    @staticmethod
    def __normalize(dist):
        min_ = dist.min()
        max_ = dist.max()
        if max_ != min_:
            std = (dist - min_) / (max_ - min_)
        else:
            std = dist / max_
        scaled = std * (1 - 0) + 0
        return scaled

    def __jensen_shannon_distance(self, p, q):
        """
        method to compute the Jenson-Shannon Distance
        between two probability distributions. 0 < distance < 1.
        """

        # convert the vectors into numpy arrays in case that they aren't
        p = self.__normalize(np.array(p))
        q = self.__normalize(np.array(q))

        m = (p + q) / 2

        # compute Jensen Shannon Divergence
        divergence = (st.entropy(p, m) + st.entropy(q, m)) / 2

        # compute the Jensen Shannon Distance
        distance = np.sqrt(divergence)

        return distance + 1e-6


class Correlations(BaseMetric):
    def __init__(
        self,
        original: pd.DataFrame,
        synthetic: pd.DataFrame,
        plot: bool,
        reports_path: str,
    ):
        super().__init__(original, synthetic, plot, reports_path)
        self.cmap = LinearSegmentedColormap.from_list(
            "rg",
            list(reversed(["#96195C", "#C13666", "#B24E89", "#9075C1", "#3F93E1", "#E8F4FF"])),
        )

    def calculate_all(self, categorical_columns: List[str], cont_columns: List[str]):
        categorical_columns = [col for col in categorical_columns if "word_count" not in col]
        cont_columns = [col for col in cont_columns if "word_count" not in col]
        for col in categorical_columns:
            map_dict = {
                k: i + 1 for i, k in enumerate(set(self.original[col]) | set(self.synthetic[col]))
            }
            self.original[col] = self.original[col].map(map_dict)
            self.synthetic[col] = self.synthetic[col].map(map_dict)

        self.original_heatmap = self.__calculate_correlations(
            self.original[categorical_columns + cont_columns].apply(
                pd.to_numeric, axis=0, errors="ignore"
            )
        )
        self.synthetic_heatmap = self.__calculate_correlations(
            self.synthetic[categorical_columns + cont_columns].apply(
                pd.to_numeric, axis=0, errors="ignore"
            )
        )
        self.corr_score = self.original_heatmap - self.synthetic_heatmap
        self.corr_score = (
            self.corr_score
            .dropna(how="all")
            .dropna(how="all", axis=1)
        )

        # check if there are any nans left in corr_score
        if self.corr_score.isna().values.any():
            # mask for NaNs in both original_heatmap and synthetic_heatmap
            nan_mask = (
                np.isnan(self.original_heatmap) &
                np.isnan(self.synthetic_heatmap)
            )

            # Set the NaN values in corr_score to 0 where both
            # original_heatmap and synthetic_heatmap have NaNs
            self.corr_score[nan_mask] = 0

        if self.plot:
            self.__create_correlation_heatmap()

        return np.median(self.corr_score)

    def __create_correlation_heatmap(self):
        plt.clf()
        # workaround for the issue not showing annotations for NaN values
        # change NaN values to 2 that is not in the range 0-1
        corr_score = self.corr_score.fillna(2)
        # set mask for NaN values
        nan_mask = corr_score == 2
        # Color for values over vmax
        self.cmap.set_over('gray')

        # annotation matrix where 'NaN' is placed where the data is NaN
        annotations = np.where(nan_mask, 'NaN', '')

        sns.set(rc={"figure.figsize": self.corr_score.shape}, font_scale=2)
        heatmap = sns.heatmap(
            corr_score,
            annot=annotations,
            cmap=self.cmap,
            vmin=0.0,
            vmax=1.0,
            center=0.5,
            square=True,
            fmt='s',
            annot_kws={'color': 'white'}
        )

        heatmap.figure.tight_layout()
        plt.savefig(
            f"{self.reports_path}/correlations_heatmap.svg",
            bbox_inches="tight",
            format="svg",
        )

    @staticmethod
    def __calculate_correlations(data):
        return abs(data.corr(method="spearman"))


class BivariateMetric(BaseMetric):
    def __init__(
        self,
        original: pd.DataFrame,
        synthetic: pd.DataFrame,
        plot: bool,
        reports_path: str,
        dataset_pickle_path: str
    ):
        super().__init__(original, synthetic, plot, reports_path)
        self.cmap = LinearSegmentedColormap.from_list(
            "rg", ["#0D5598", "#3E92E0", "#E8F4FF"]
        )
        self.dataset = fetch_config(dataset_pickle_path)
        nan_labels_dict = self.dataset.nan_labels_dict
        na_values = [f'\'{value}\'' for value in self.dataset.format.get("na_values", [])]
        if na_values:
            logger.info(
                f"During data loading the 'format' section was used, therefore the values - "
                f"{', '.join(na_values)} have been replaced with 'nan', impacting bivariate plots"
            )
        self.missing_values: Dict[str, str] = (
            {col: nan_labels_dict.get(col, "nan") for col in self.dataset.order_of_columns}
            if nan_labels_dict
            else {}
        )

    @staticmethod
    def _format_date_labels(heatmap_orig_data, heatmap_synthetic_data, axis):
        heatmap_orig, x_tick_labels_orig, y_tick_labels_orig = heatmap_orig_data
        heatmap_synth, x_tick_labels_synth, y_tick_labels_synth = heatmap_synthetic_data
        if axis == "y":
            y_tick_labels_orig = [timestamp_to_datetime(i) for i in y_tick_labels_orig]
            y_tick_labels_synth = [timestamp_to_datetime(i) for i in y_tick_labels_synth]
        else:
            x_tick_labels_orig = [timestamp_to_datetime(i) for i in x_tick_labels_orig]
            x_tick_labels_synth = [timestamp_to_datetime(i) for i in x_tick_labels_synth]
        return (heatmap_orig, x_tick_labels_orig, y_tick_labels_orig), (
            heatmap_synth,
            x_tick_labels_synth,
            y_tick_labels_synth,
        )

    def calculate_all(
        self,
        cont_columns: List[str],
        categorical_columns: List[str],
        date_columns: Optional[List[str]],
        num_not_na_cont_ticks: int = 10,
        max_num_combinations: int = 100,
    ):
        self.date_columns = date_columns if date_columns else []
        self.num_not_na_ticks = num_not_na_cont_ticks
        fetched_columns = set(cont_columns) | set(categorical_columns)
        excluded_cols = {
            col for col in fetched_columns if col.endswith(("_word_count", "_char_len"))
        }
        all_columns = fetched_columns - excluded_cols
        all_column_pairs = list(combinations(all_columns, 2))
        column_pairs = random.sample(
            all_column_pairs, min(max_num_combinations, len(all_column_pairs))
        )
        bi_imgs = {}
        for i, (first_col, second_col) in tqdm.tqdm(
            iterable=enumerate(column_pairs),
            desc="Generating bivariate distributions...",
            total=len(column_pairs),
        ):
            plt.clf()
            fig, self._axes = plt.subplots(
                1, 2, figsize=(30, 15), gridspec_kw={"width_ratios": [7, 8.7]}
            )
            if first_col in cont_columns:
                if second_col in cont_columns:
                    (
                        heatmap_orig_data,
                        heatmap_synthetic_data,
                    ) = self._calculate_pair_continuous_vs_continuous(
                        y_col=second_col, x_col=first_col  # x is first, y is second
                    )
                elif second_col in categorical_columns:
                    (
                        heatmap_orig_data,
                        heatmap_synthetic_data,
                    ) = self._calculate_pair_continuous_vs_categ(
                        cont_col=first_col,
                        categorical_col=second_col,  # x is second, y is first
                    )

                    heatmap_orig_data = (
                        np.transpose(heatmap_orig_data[0]),
                        heatmap_orig_data[2],
                        heatmap_orig_data[1],
                    )
                    heatmap_synthetic_data = (
                        np.transpose(heatmap_synthetic_data[0]),
                        heatmap_synthetic_data[2],
                        heatmap_synthetic_data[1],
                    )  # x is first, y is second
            elif first_col in categorical_columns:
                if second_col in cont_columns:
                    (
                        heatmap_orig_data,
                        heatmap_synthetic_data,
                    ) = self._calculate_pair_continuous_vs_categ(
                        cont_col=second_col,
                        categorical_col=first_col,  # x is first, y is second
                    )
                elif second_col in categorical_columns:
                    (
                        heatmap_orig_data,
                        heatmap_synthetic_data,
                    ) = self._calculate_pair_categ_vs_categ(y_col=second_col, x_col=first_col)

            heatmap_min, heatmap_max = self.get_common_min_max(
                heatmap_orig_data[0], heatmap_synthetic_data[0]
            )

            if first_col in self.date_columns:
                heatmap_orig_data, heatmap_synthetic_data = self._format_date_labels(
                    heatmap_orig_data, heatmap_synthetic_data, "x"
                )
            if second_col in self.date_columns:
                heatmap_orig_data, heatmap_synthetic_data = self._format_date_labels(
                    heatmap_orig_data, heatmap_synthetic_data, "y"
                )

            self._plot_heatmap(heatmap_orig_data, plt_index=0,
                               vrange=(heatmap_min, heatmap_max),
                               features=(first_col, second_col),
                               cbar=False)

            self._plot_heatmap(heatmap_synthetic_data, plt_index=1,
                               vrange=(heatmap_min, heatmap_max),
                               features=(first_col, second_col),
                               cbar=True)
            # first_col is x axis, second_col is y axis
            title = f"{first_col} vs. {second_col}"
            path_to_image = (
                f"{self.reports_path}/bivariate_{slugify(first_col)}_{slugify(second_col)}.svg"
            )
            bi_imgs[title] = path_to_image
            plt.savefig(path_to_image, format="svg")
        return bi_imgs

    @staticmethod
    def get_common_min_max(original, synthetic):
        original_flat = [c for row in original for c in row]
        synthetic_flat = [c for row in synthetic for c in row]
        vmax = max(max(original_flat), max(synthetic_flat))
        vmin = min(min(original_flat), min(synthetic_flat))
        return vmin, vmax

    @staticmethod
    def __format_float_tick_labels(labels: List, nan_label: str = "nan") -> List:
        labels = [nan_label if pd.isna(label) else label for label in labels]
        if all([isinstance(i, float) for i in labels]) and (
            max(labels) > 1e5 or min(labels) < 1e-03
        ):
            return [f"{label:.4e}" for label in labels]
        if all([isinstance(i, float) for i in labels]):
            return [f"{round(i, 4)}" for i in labels]
        return labels

    def _plot_heatmap(
        self,
        heatmap_data: List,
        plt_index: int,
        vrange: tuple[float],
        features: tuple[str],
        cbar=True,
    ):
        vmin, vmax = vrange
        xfeature, yfeature = features
        ax = self._axes.flat[plt_index]
        ax.tick_params(labelsize=14)
        heatmap, x_tick_labels, y_tick_labels = heatmap_data
        x_tick_labels = self.__format_float_tick_labels(
            x_tick_labels,
            self.missing_values.get(xfeature, "nan")
        )
        y_tick_labels = self.__format_float_tick_labels(
            y_tick_labels,
            self.missing_values.get(yfeature, "nan")
        )
        ax = sns.heatmap(
            heatmap,
            xticklabels=x_tick_labels,
            yticklabels=y_tick_labels,
            ax=ax,
            vmin=vmin,
            vmax=vmax,
            cmap=self.cmap,
            cbar=cbar,
        )
        if cbar:
            cbar = ax.collections[0].colorbar
            cbar.ax.tick_params(axis="y", labelsize=20)

    def _get_continuous_ticks(self, col_name: str):
        original_col_values = self.original[col_name].dropna().values
        synthetic_col_values = self.original[col_name].dropna().values
        col_values = np.concatenate((original_col_values, synthetic_col_values))
        return np.percentile(col_values, list(range(0, 101, self.num_not_na_ticks)))

    def _get_categorical_ticks(self, col_name: str, ticks_count: int = 50):
        """
        Select the ticks for a categorical column
        """
        min_ticks = int(ticks_count * 0.2)
        max_ticks = int(ticks_count * 0.2)
        other_ticks = ticks_count - min_ticks - max_ticks

        categ_ticks = sorted(
            set(
                set(self.original[col_name].fillna("?"))
                | set(self.synthetic[col_name].fillna("?"))
            )
        )
        # Select the first min_ticks elements
        selected_min_ticks = categ_ticks[:min_ticks]

        # Select the last max_ticks elements
        selected_max_ticks = categ_ticks[-max_ticks:]

        # Remove the already selected elements from categ_ticks
        remaining_ticks = categ_ticks[min_ticks:-max_ticks]

        # Randomly select other_ticks elements from the remaining list
        selected_other_ticks = random.sample(
            remaining_ticks, min(other_ticks, len(remaining_ticks))
        )

        return sorted(
            set(
                selected_min_ticks + selected_other_ticks + selected_max_ticks
                )
            )

    @staticmethod
    def _format_categorical_labels(labels):
        return [
            str(label) if len(str(label)) < 15 else str(label[:12]) + "..." for label in labels
        ]

    @staticmethod
    def _smooth(dist):
        for i in range(1, len(dist)):
            if dist[i - 1] == dist[i]:
                dist[i:] += 2
        dist.sort()
        return dist

    def _calculate_pair_categ_vs_categ(self, y_col: str, x_col: str):
        def calc_heatmap_data(
            df: pd.DataFrame,
            y_col: str,
            x_col: str,
            xticks: List[str],
            yticks: List[float],
        ):
            heatmap = []
            for y_val in yticks:
                heatmap_row = []
                for x_val in xticks:
                    count = df[df[y_col].eq(y_val) & df[x_col].eq(x_val)].shape[0] / len(df)
                    heatmap_row.append(count)
                heatmap.append(heatmap_row)
            return (
                heatmap,
                self._format_categorical_labels(xtick),
                self._format_categorical_labels(ytick),
            )

        xtick = self._get_categorical_ticks(x_col)
        ytick = self._get_categorical_ticks(y_col)

        original_heatmap = calc_heatmap_data(self.original, y_col, x_col, xtick, ytick)
        synthetic_heatmap = calc_heatmap_data(self.synthetic, y_col, x_col, xtick, ytick)
        return original_heatmap, synthetic_heatmap

    def _calculate_pair_continuous_vs_categ(self, cont_col: str, categorical_col: str):
        def calc_heatmap_data(
            df: pd.DataFrame,
            cont_col: str,
            categorical_col: str,
            xticks: List[float],
            yticks: List[str],
        ):
            heatmap = []
            cont_has_nan = df[cont_col].isnull().values.any()
            cont_ticks = self.num_not_na_ticks + 1 if cont_has_nan else self.num_not_na_ticks
            for i in range(cont_ticks):
                heatmap_row = []
                if i == 0:
                    x_index = (df[cont_col] >= xticks[i]) & (df[cont_col] <= xticks[i + 1])
                elif i < self.num_not_na_ticks:
                    x_index = (df[cont_col] > xticks[i]) & (df[cont_col] <= xticks[i + 1])
                else:
                    x_index = df[cont_col].isna()

                for val in yticks:
                    count = df[x_index & df[categorical_col].eq(val)].shape[0] / len(df)
                    heatmap_row.append(count)
                heatmap.append(heatmap_row)

            if cont_has_nan:
                xticks = np.append(xticks, np.nan)

            return heatmap, self._format_categorical_labels(yticks), xticks[1:]

        xtick = self._smooth(self._get_continuous_ticks(cont_col))
        ytick = self._get_categorical_ticks(categorical_col)

        original_heatmap = calc_heatmap_data(
            self.original,
            cont_col,
            categorical_col,
            xtick,
            ytick
        )
        synthetic_heatmap = calc_heatmap_data(
            self.synthetic,
            cont_col,
            categorical_col,
            xtick,
            ytick
        )

        return original_heatmap, synthetic_heatmap

    def _calculate_pair_continuous_vs_continuous(self, y_col: str, x_col: str):
        def calc_heatmap_data(
            df: pd.DataFrame,
            y_col: str,
            x_col: str,
            xticks: List[float],
            yticks: List[float],
        ):
            heatmap = []
            y_has_nan = df[y_col].isnull().values.any()
            x_has_nan = df[x_col].isnull().values.any()
            y_ticks_num = self.num_not_na_ticks + 1 if y_has_nan else self.num_not_na_ticks
            x_ticks_num = self.num_not_na_ticks + 1 if x_has_nan else self.num_not_na_ticks

            for i in range(y_ticks_num):
                heatmap_row = []
                if i == 0:
                    y_index = (df[y_col] >= yticks[i]) & (df[y_col] <= yticks[i + 1])
                elif i < self.num_not_na_ticks:
                    y_index = (df[y_col] > yticks[i]) & (df[y_col] <= yticks[i + 1])
                else:
                    y_index = df[y_col].isna()

                for j in range(x_ticks_num):
                    if j == 0:
                        x_index = (df[x_col] >= xticks[j]) & (df[x_col] <= xticks[j + 1])
                    elif j < self.num_not_na_ticks:
                        x_index = (df[x_col] > xticks[j]) & (df[x_col] <= xticks[j + 1])
                    else:
                        x_index = df[x_col].isna()

                    count = df[y_index & x_index].shape[0] / len(df)
                    heatmap_row.append(count)

                heatmap.append(heatmap_row)

            if x_has_nan:
                xticks = np.append(xticks, np.nan)
            if y_has_nan:
                yticks = np.append(yticks, np.nan)
            return heatmap, xticks[1:], yticks[1:]

        ytick = self._smooth(self._get_continuous_ticks(y_col))
        xtick = self._smooth(self._get_continuous_ticks(x_col))

        original_heatmap = calc_heatmap_data(self.original, y_col, x_col, xtick, ytick)
        synthetic_heatmap = calc_heatmap_data(self.synthetic, y_col, x_col, xtick, ytick)

        return original_heatmap, synthetic_heatmap


class UnivariateMetric(BaseMetric):
    def __init__(
        self,
        original: pd.DataFrame,
        synthetic: pd.DataFrame,
        plot: bool,
        reports_path: str,
    ):
        super().__init__(original, synthetic, plot, reports_path)

    def _calculate(self, column):
        pass

    def calculate_all(
        self,
        cont_columns: List[str],
        categorical_columns: List[str],
        date_columns: Optional[List[str]],
        print_nan: bool = False,
    ):
        cont_columns = list(cont_columns)
        excluded_cols = {
            col for col in cont_columns if "word_count" if col.endswith("_word_count")
        }
        cont_columns = set(cont_columns) - excluded_cols
        excluded_cols = {
            col for col in categorical_columns if "word_count" if col.endswith("_word_count")
        }
        categorical_columns = set(categorical_columns) - excluded_cols
        images = {}
        uni_categ_images = {}
        for col in cont_columns:
            images.update(
                self.__calc_continuous(column=col, print_nan=print_nan, isdate=col in date_columns)
            )
        for col in categorical_columns:
            uni_categ_images.update(self.__calc_categ(col))
        images.update(uni_categ_images)
        return images

    @staticmethod
    def _get_ratio_counts(ratio_counts, count: int = 30) -> Dict:
        """
        Select the most common, least common and some random items between them
        from the ratio_counts
        """
        ratio_counts = Counter(ratio_counts)

        most_least_count = int(count * 0.2)
        other_items = count - 2 * most_least_count

        most_common_items = ratio_counts.most_common(most_least_count)
        least_common_items = ratio_counts.most_common()[: -most_least_count - 1: -1]
        between_items = ratio_counts.most_common()[most_least_count:-most_least_count]
        selected_between_items = random.sample(between_items, min(other_items, len(between_items)))

        updated_ratio_counts = dict(
            most_common_items + selected_between_items + least_common_items
        )

        return updated_ratio_counts

    def __calc_categ(self, column):
        def plot_dist(column_data, sort=True, full_set=None):
            counts = Counter(column_data)
            if full_set is not None:
                absent_keys = full_set - set(counts.keys())
                if len(absent_keys) > 0:
                    for k in absent_keys:
                        counts[k] = 0

            if sort:
                sorted_keys = sorted(counts, key=lambda x: counts[x])
                counts = {i: counts[i] for i in sorted_keys}

            size = len(column_data)
            counts = {key: (x / size * 100) for key, x in counts.items()}
            return counts

        def sanitize_labels(label):
            return re.sub(r"\$|\^", "", label)

        original_column = self.original[column]
        synthetic_column = self.synthetic[column]
        full_values_set = set(original_column.values) | set(synthetic_column.values)

        original_ratio_counts = plot_dist(original_column, True, full_values_set)
        original_ratio_counts = self._get_ratio_counts(original_ratio_counts)
        original_labels = list(original_ratio_counts.keys())
        original_ratio = list(original_ratio_counts.values())

        synthetic_ratio_counts = plot_dist(synthetic_column, False, full_values_set)
        synthetic_ratio = [
            synthetic_ratio_counts[label]
            for label in original_labels
            if synthetic_ratio_counts.get(label) is not None
        ]

        uni_images = {}

        if self.plot:
            fig = plt.figure(figsize=(9.5, 6.5))

            width = 0.35
            x = np.arange(len(original_labels))
            plt.bar(
                x - width / 2,
                original_ratio,
                width=width,
                label="Original",
                color="#3F93E1",
            )
            plt.bar(
                x + width / 2,
                synthetic_ratio,
                width=width,
                label="Synthetic",
                color="#FF9C54",
            )
            ax = plt.gca()
            ax.spines[["top", "right", "left", "bottom"]].set_color("#E5E9EB")
            ax.yaxis.grid(color="#E5E9EB", linestyle="-", linewidth=0.5)
            ax.set_axisbelow(True)
            ax.set_facecolor("#FFFFFF")
            ax.set_xticks(x)
            ax.set_xticklabels(
                [
                    str(sanitize_labels(label[:30])) + "..."
                    if len(str(label)) > 33
                    else sanitize_labels(str(label))
                    for label in original_labels
                ]
            )
            ax.tick_params(axis="both", which="major", labelsize=9)
            fig.autofmt_xdate()
            ax.set_xlabel("category", fontsize=9)
            ax.set_ylabel("percents", fontsize=9)
            plt.legend(
                ["original", "synthetic"],
                loc="upper left",
                fontsize=9,
                bbox_to_anchor=(0, 1.07),
                ncol=2,
                frameon=False,
            )
            if self.reports_path:
                path_to_image = f"{self.reports_path}/univariate_{slugify(column)}.svg"
                plt.savefig(path_to_image, bbox_inches="tight", format="svg")
                uni_images[column] = path_to_image
        return uni_images

    def __calc_continuous(
            self,
            column: str,
            isdate: bool,
            print_nan: bool = False,
            kurtosis_threshold: float = KURTOSIS_THRESHOLD
            ):
        original_nan_count = self.original[column].isna().sum()
        synthetic_nan_count = self.synthetic[column].isna().sum()
        original_unique_count = self.original[column].nunique()
        synthetic_unique_count = self.synthetic[column].nunique()

        uni_images = {}

        if self.plot and original_unique_count > 1 and synthetic_unique_count > 1:
            fig, ax = plt.subplots(figsize=(8, 6.5))

            # Kernel Density Estimation plot using Seaborn
            sns.kdeplot(data=self.original, x=column, color="#3F93E1", linewidth=2, ax=ax)
            sns.kdeplot(data=self.synthetic, x=column, color="#FF9C54", linewidth=2, ax=ax)

            ax.set_xlabel("value", fontsize=9)
            ax.set_ylabel("density", fontsize=9)
            ax.tick_params(axis="both", which="major", labelsize=9)
            ax.ticklabel_format(axis="x", style="plain")
            fmt = ax.yaxis.get_major_formatter()
            fmt.set_useMathText(True)
            ax.yaxis.offsetText.set_visible(False)
            ax.yaxis.set_label_text(f"density {fmt.get_offset()}")
            ax.yaxis.get_offset_text().set_fontsize(9)
            plt.xticks(rotation=45)
            ax.spines[["top", "right", "left", "bottom"]].set_color("#E5E9EB")
            ax.grid(color="#E5E9EB", linestyle="-", linewidth=0.5)
            ax.set_axisbelow(True)
            ax.set_facecolor("#FFFFFF")
            plt.legend(
                ["original", "synthetic"],
                loc="upper left",
                fontsize=9,
                bbox_to_anchor=(0, 1.07),
                ncol=2,
                frameon=False,
            )
            if isdate:
                lower_x, upper_x = ax.dataLim._points[:, 0]
                len_x_labels = len(ax.get_xticklabels())

                x_ticks = np.linspace(lower_x, upper_x, len_x_labels)
                x_ticks = [timestamp_to_datetime(i) for i in x_ticks]
                ax.set_xticklabels(x_ticks, rotation=45, ha="right")
            if self.reports_path:
                path_to_image = f"{self.reports_path}/univariate_{slugify(column)}.svg"
                plt.savefig(path_to_image, bbox_inches="tight", format="svg")
                uni_images[column] = path_to_image
        if print_nan:
            logger.info(f"Number of original NaN values in {column}: {original_nan_count}")
            logger.info(f"Number of synthetic NaN values in {column}: {synthetic_nan_count}")

        kurt = kurtosis(self.original[column], nan_policy='omit')
        if kurt > kurtosis_threshold:
            outliers_metrics = self._calculate_outliers_metrics(column)
            self._log_outliers_analysis(outliers_metrics, column)

        return uni_images

    def _calculate_outliers_metrics(self, column):
        """
        Calculate various outliers-related metrics for a given column
        in the original and synthetic datasets.

        Parameters:
            column (str): The name of the column to analyze.

        Returns:
            dict: A dictionary containing kurtosis, top value ratios,
            max value ratios, outlier ratios, extreme outlier ratios
            and skewness for both original and synthetic data.
        """
        metrics = {}

        # Kurtosis
        original_kurtosis = kurtosis(self.original[column], nan_policy='omit')
        synthetic_kurtosis = kurtosis(self.synthetic[column], nan_policy='omit')
        kurtosis_ratio = self._calculate_ratio(
            original_kurtosis, synthetic_kurtosis
        )
        metrics["kurtosis_original"] = original_kurtosis
        metrics["kurtosis_synthetic"] = synthetic_kurtosis
        metrics["kurtosis_ratio"] = kurtosis_ratio

        # Top1 value ratio
        original_top1_value, original_top1_ratio = next(
            self.original[column].value_counts(normalize=True).items()
        )
        synthetic_top1_value, synthetic_top1_ratio = next(
            self.synthetic[column].value_counts(normalize=True).items()
        )
        top1_ratio_diff = abs(synthetic_top1_ratio - original_top1_ratio)
        metrics["top1_ratio_original"] = original_top1_ratio
        metrics["top1_ratio_synthetic"] = synthetic_top1_ratio
        metrics["top1_ratio_diff"] = top1_ratio_diff

        # Max value
        original_max_value = self.original[column].max()
        synthetic_max_value = self.synthetic[column].max()
        max_value_ratio = self._calculate_ratio(
            original_max_value, synthetic_max_value
        )
        metrics["max_value_original"] = original_max_value
        metrics["max_value_synthetic"] = synthetic_max_value
        metrics["max_value_ratio"] = max_value_ratio

        # Min value
        original_min_value = self.original[column].min()
        synthetic_min_value = self.synthetic[column].min()
        min_value_ratio = self._calculate_ratio(
            original_min_value, synthetic_min_value
        )
        metrics["min_value_original"] = original_min_value
        metrics["min_value_synthetic"] = synthetic_min_value
        metrics["min_value_ratio"] = min_value_ratio

        # Outlier ratio using IQR
        original_outlier_ratio = self.outlier_ratio_iqr(
            self.original[column].dropna()
        )
        synthetic_outlier_ratio = self.outlier_ratio_iqr(
            self.synthetic[column].dropna()
        )
        outlier_ratio_diff = (
            abs(synthetic_outlier_ratio - original_outlier_ratio)
        )
        metrics["outlier_ratio_original"] = original_outlier_ratio
        metrics["outlier_ratio_synthetic"] = synthetic_outlier_ratio
        metrics["outlier_ratio_diff"] = outlier_ratio_diff

        # Extreme outlier ratio using IQR with factor 10
        original_extr_outlier_ratio = self.outlier_ratio_iqr(
            self.original[column].dropna(), factor=10
        )
        synthetic_extr_outlier_ratio = self.outlier_ratio_iqr(
            self.synthetic[column].dropna(), factor=10
        )
        extr_outlier_ratio_diff = (
            abs(synthetic_extr_outlier_ratio - original_extr_outlier_ratio)
        )
        metrics["extr_outlier_ratio_original"] = original_extr_outlier_ratio
        metrics["extr_outlier_ratio_synthetic"] = synthetic_extr_outlier_ratio
        metrics["extr_outlier_ratio_diff"] = extr_outlier_ratio_diff
        # Skewness
        original_skewness = skew(self.original[column], nan_policy='omit')
        synthetic_skewness = skew(self.synthetic[column], nan_policy='omit')
        skewness_ratio = self._calculate_ratio(
            original_skewness, synthetic_skewness
        )
        metrics["skewness_original"] = original_skewness
        metrics["skewness_synthetic"] = synthetic_skewness
        metrics["skewness_ratio"] = skewness_ratio
        return metrics

    def _log_outliers_analysis(self, metrics: dict, column: str):
        logger.trace(
            f"\nOutliers analysis for '{column}':"
            f"\n Kurtosis, '{column}':"
            f"\n  - Original: {metrics['kurtosis_original']:.2f}"
            f"\n  - Synthetic: {metrics['kurtosis_synthetic']:.2f}"
            f"\n  - Ratio (synthetic/original): "
            f"{metrics['kurtosis_ratio']:.2f}"
            f"\n"
            f"\nTop1 Value Ratio, '{column}':"
            f"\n  - Original: {metrics['top1_ratio_original']:.4f}"
            f"\n  - Synthetic: {metrics['top1_ratio_synthetic']:.4f}"
            f"\n  - Absolute Difference: "
            f"{metrics['top1_ratio_diff']:.4f}"
            f"\n"
            f"\nOutliers Ratio (IQR), '{column}':"
            f"\n  - Original: "
            f"{metrics['outlier_ratio_original']:.2%}"
            f"\n  - Synthetic: "
            f"{metrics['outlier_ratio_synthetic']:.2%}"
            f"\n  - Absolute Difference: "
            f"{metrics['outlier_ratio_diff']:.2%}"
            f"\n"
            f"\nExtreme Outliers Ratio (IQR, factor=10), '{column}':"
            f"\n  - Original: "
            f"{metrics['extr_outlier_ratio_original']:.2%}"
            f"\n  - Synthetic: "
            f"{metrics['extr_outlier_ratio_synthetic']:.2%}"
            f"\n  - Absolute Difference: "
            f"{metrics['extr_outlier_ratio_diff']:.2%}"
            f"\n"
            f"\nMax Value, '{column}':"
            f"\n  - Original: {metrics['max_value_original']:.2f}"
            f"\n  - Synthetic: {metrics['max_value_synthetic']:.2f}"
            f"\n  - Ratio (synthetic/original): "
            f"{metrics['max_value_ratio']:.2f}"
            f"\n"
            f"\nMin Value, '{column}':"
            f"\n  - Original: {metrics['min_value_original']:.2f}"
            f"\n  - Synthetic: {metrics['min_value_synthetic']:.2f}"
            f"\n  - Ratio (synthetic/original): "
            f"{metrics['min_value_ratio']:.2f}"
            f"\n"
            f"\nSkewness, '{column}':"
            f"\n  - Original: {metrics['skewness_original']:.2f}"
            f"\n  - Synthetic: {metrics['skewness_synthetic']:.2f}"
            f"\n  - Ratio (synthetic/original): "
            f"{metrics['skewness_ratio']:.2f}"
            f"\n"
        )

    def _calculate_ratio(
            self, original_metric, synthetic_metric, epsilon=1e-10
    ):
        """
        Calculate the ratio of two values
        with proper handling of near-zero values.

        Args:
            original_metric: Value from original data
            synthetic_metric: Value from synthetic data
            epsilon: Threshold for considering values as effectively zero

        Returns:
            float: Ratio where:
            - 1.0 if both metrics are effectively zero (< epsilon)
            - synthetic/epsilon if only original is effectively zero
            - regular ratio otherwise
        """
        # Check if values are effectively zero
        original_is_zero = abs(original_metric) < epsilon
        synthetic_is_zero = abs(synthetic_metric) < epsilon

        if original_is_zero and synthetic_is_zero:
            return 1.0

        if original_is_zero and not synthetic_is_zero:
            return abs(synthetic_metric / epsilon)

        return abs(synthetic_metric / original_metric)

    @staticmethod
    def outlier_ratio_iqr(column: pd.Series, factor=1.5):
        Q1 = column.quantile(0.25)
        Q3 = column.quantile(0.75)
        IQR = Q3 - Q1
        outlier_mask = (column < (Q1 - factor * IQR)) | (column > (Q3 + factor * IQR))
        return outlier_mask.mean()


class Clustering(BaseMetric):
    def __init__(
        self,
        original: pd.DataFrame,
        synthetic: pd.DataFrame,
        plot: bool,
        reports_path: str,
    ):
        super().__init__(original, synthetic, plot, reports_path)

    def calculate_all(self, categorical_columns: List[str], cont_columns: List[str]):
        for col in categorical_columns:
            map_dict = {
                k: i + 1 for i, k in enumerate(set(self.original[col]) | set(self.synthetic[col]))
            }
            self.original[col] = self.original[col].map(map_dict)
            self.synthetic[col] = self.synthetic[col].map(map_dict)

        # Perform clustering of original dataset
        # to get optimal number of clusters
        original_for_clustering = self.original[
            cont_columns + categorical_columns
            ].dropna()
        original_transformed = self.__preprocess_data(original_for_clustering)

        if len(original_transformed) == 0:
            logger.warning(
                "No clustering metric will be formed due to empty DataFrame"
            )
            return None

        optimal_clust_num = self.__get_optimal_number_of_clusters(
            original_transformed
            )

        logger.trace(f"Optimal number of clusters for "
                     f"original dataset: {optimal_clust_num}"
                     )

        row_limit = min(len(self.original), len(self.synthetic))

        # TODO check whether random_state affects the results
        self.merged = (
            pd.concat(
                [
                    self.original[cont_columns + categorical_columns].sample(
                        row_limit,
                        random_state=10
                    ),
                    self.synthetic[cont_columns + categorical_columns].sample(
                        row_limit,
                        random_state=10
                    ),
                ],
                keys=["original", "synthetic"],
            )
            .dropna()
            .reset_index()
        )

        if len(self.merged) == 0:
            logger.warning(
                "No clustering metric will be formed due to empty DataFrame"
            )
            return None
        self.merged_transformed = self.__preprocess_data(self.merged)

        statistics = self.__calculate_clusters(optimal_clust_num)
        statistics.columns = ["cluster", "dataset", "count"]

        diversity_scores = statistics.groupby('cluster').apply(
            self.calculate_diversity
            )
        mean_score = diversity_scores.mean()

        if self.plot:
            plt.clf()
            sns.set(font_scale=3)
            barplot = sns.barplot(
                data=statistics,
                x="cluster",
                y="count",
                hue="dataset",
                palette={"synthetic": "#ff9c54", "original": "#3f93e1"},
                saturation=1,
            )
            barplot.set_facecolor("#FFFFFF")
            ax = plt.gca()
            ax.spines[["top", "right", "left", "bottom"]].set_color("#E5E9EB")
            ax.yaxis.grid(color="#E5E9EB", linestyle="-", linewidth=1)
            original_label = Patch(color="#3f93e1", label="original")
            synthetic_label = Patch(color="#ff9c54", label="synthetic")
            plt.legend(
                handles=[original_label, synthetic_label],
                loc="upper left",
                bbox_to_anchor=(0, 1.09),
                ncol=2,
                frameon=False,
            )
            plt.savefig(
                f"{self.reports_path}/clusters_barplot.svg",
                bbox_inches="tight",
                format="svg",
            )
        return mean_score

    @staticmethod
    def calculate_diversity(statistics):
        """
        Calculate the diversity score for each cluster
        from collected statistics.
        If in cluster only one dataset is present, return 0.
        """
        if statistics['dataset'].nunique() == 2:
            return min(statistics['count']) / max(statistics['count'])
        else:
            return 0

    def __get_optimal_number_of_clusters(self, dataset):
        """
        Calculate the optimal number of clusters using Davies-Bouldin score
        """
        davies_bouldin_scores = []
        max_clusters = min(10, len(dataset))

        for i in range(2, max_clusters):
            clusters = KMeans(n_clusters=i, random_state=10).fit(
                dataset
                )
            labels = clusters.labels_
            score = davies_bouldin_score(dataset, labels)
            davies_bouldin_scores.append(score)

        # Get number of clusters with the lowest Davies-Bouldin score
        # +2 because the range starts from 2
        optimal_clusters = np.argmin(davies_bouldin_scores) + 2

        return optimal_clusters

    def __preprocess_data(self, dataset):
        transformed_dataset = dataset.apply(
            pd.to_numeric, axis=0, errors="ignore"
        ).select_dtypes(include="number")
        scaler = MinMaxScaler()
        transformed_dataset = scaler.fit_transform(transformed_dataset)

        return transformed_dataset

    def __calculate_clusters(self, n):
        clusters = KMeans(n_clusters=n, random_state=10).fit(
            self.merged_transformed
        )
        labels = clusters.labels_
        rows_labels = pd.DataFrame(
            {"origin": self.merged["level_0"], "cluster": labels}
        )

        return rows_labels.groupby(["cluster", "origin"]).size().reset_index()


class Utility(BaseMetric):
    def __init__(
        self,
        original: pd.DataFrame,
        synthetic: pd.DataFrame,
        plot: bool,
        reports_path: str,
        sample_size: int = 100000,
    ):
        super().__init__(original, synthetic, plot, reports_path)

        self.sample_size = sample_size

    @staticmethod
    def check_empty_df(df: pd.DataFrame, df_type: Literal["original", "synthetic"]) -> bool:
        """
        Check if the dataframe is empty after dropping rows with missing values
        """
        if df.empty:
            logger.warning(
                f"Utility metric calculation is skipped: the {df_type} dataframe is empty "
                "after dropping rows with missing values (dropna() function is applied)"
            )
            return True
        return False

    def calculate_all(self, categorical_columns: List[str], cont_columns: List[str]):
        logger.info("Calculating utility metric")

        for col in categorical_columns:
            map_dict = {
                k: i + 1 for i, k in enumerate(
                    set(self.original[col]) | set(self.synthetic[col])
                )
            }
            self.original[col] = self.original[col].map(map_dict)
            self.synthetic[col] = self.synthetic[col].map(map_dict)

        self.original = self.original[cont_columns + categorical_columns].apply(
            pd.to_numeric, axis=0, errors="ignore"
        )
        self.synthetic = self.synthetic[cont_columns + categorical_columns].apply(
            pd.to_numeric, axis=0, errors="ignore"
        )
        self.original = self.original.select_dtypes(include="number").dropna()
        self.synthetic = self.synthetic.select_dtypes(include="number").dropna()

        if self.check_empty_df(self.original, "original"):
            return
        if self.check_empty_df(self.synthetic, "synthetic"):
            return

        self.synthetic = self.synthetic[self.original.columns]

        excluded_cols = [
            col
            for col in categorical_columns + cont_columns
            if self.original[col].nunique() < 2 or col.endswith(("_word_count", "_char_len"))
        ]
        binary_cols = [
            col
            for col in categorical_columns
            if self.original[col].nunique() == 2 and col not in excluded_cols
        ]
        cont_cols = [
            col
            for col in cont_columns
            if col not in binary_cols and col not in excluded_cols
        ]
        categorical_cols = [
            col
            for col in categorical_columns
            if col not in binary_cols and col not in excluded_cols
        ]

        # Check if the datasets after dropna() have more than
        # self.sample_size rows and stratified sampling is needed
        self.is_big_original_data_after_drop_na = (
            len(self.original) > self.sample_size
        )
        self.is_big_synthetic_data_after_drop_na = (
            len(self.synthetic) > self.sample_size
        )

        log_debug_msg = (
            "There are {} records left "
            "in {} data after dropping rows with null values "
            "in numeric columns. "
            "Creating stratified samples of size {} "
            "to calculate utility metric"
        )

        log_info_msg = (
            "There are more than {} records "
            "in {} data. "
            "Creating stratified samples of size {} "
            "to calculate utility metric"
        )

        if self.is_big_original_data_after_drop_na:
            logger.debug(log_debug_msg.format(
                len(self.original),
                'original',
                self.sample_size
                )
            )
            logger.info(
                log_info_msg.format(
                    self.sample_size,
                    'original',
                    self.sample_size
                )
            )

        if self.is_big_synthetic_data_after_drop_na:
            logger.debug(log_debug_msg.format(
                len(self.synthetic),
                'synthetic',
                self.sample_size
                )
            )
            logger.info(
                log_info_msg.format(
                    self.sample_size,
                    'synthetic',
                    self.sample_size
                )
            )

        (
            best_categ,
            score_categ,
            synth_score_categ
        ) = self.__create_multi_class_models(categorical_cols)
        (
            best_binary,
            score_binary,
            synth_score_binary,
        ) = self.__create_binary_class_models(binary_cols)
        (
            best_regres,
            score_regres,
            synth_regres_score
        ) = self.__create_regression_models(cont_cols)

        result = pd.DataFrame(
            {
                "original": [
                    round(score_binary, 3) if best_binary is not None else np.nan,
                    round(score_categ, 3) if best_categ is not None else np.nan,
                    # Do this to avoid zero division
                    round(max(score_regres, 1e-10), 3) if best_regres is not None else np.nan,
                ],
                "synthetic": [
                    round(synth_score_binary, 3) if best_binary is not None else np.nan,
                    round(synth_score_categ, 3) if best_categ is not None else np.nan,
                    abs(round(max(0, synth_regres_score), 3))
                    if best_regres is not None else np.nan,
                ],
                "synth_to_orig_ratio": [
                    round(synth_score_binary / score_binary, 3)
                    if best_binary is not None and score_binary != 0 else (
                        0 if score_binary == 0 else np.nan
                        ),
                    round(synth_score_categ / score_categ, 3)
                    if best_categ is not None and score_categ != 0 else (
                        0 if score_categ == 0 else np.nan
                        ),
                    abs(round(max(0, synth_regres_score) / score_regres, 3))
                    if best_regres is not None and score_regres != 0 else (
                        0 if score_regres == 0 else np.nan
                        ),
                ],
                "type": [
                    "Binary (" + f"{best_binary})" if best_binary is not None else "" + ")",
                    "Multiclass (" + f"{best_categ})" if best_categ is not None else "" + ")",
                    "Regression (" + f"{best_regres})" if best_regres is not None else "" + ")",
                ],
            }
        )
        result = pd.melt(result.dropna(), id_vars=["type", "synth_to_orig_ratio"])
        result.columns = ["Type", "Synth to orig ratio", "Origin", "Model score"]
        result = result.sort_values("Type").reset_index(drop=True)

        if self.plot:
            if result.empty:
                logger.info("No data to provide utility barplot")
            else:
                plt.clf()
                sns.set(font_scale=3)
                barplot = sns.barplot(
                    data=result,
                    x="Type",
                    y="Model score",
                    hue="Origin",
                    palette={"synthetic": "#FF9C54", "original": "#3F93E1"},
                    saturation=1,
                )
                barplot.set_facecolor("#FFFFFF")
                ax = plt.gca()
                ax.spines[["top", "right", "left", "bottom"]].set_color("#E5E9EB")
                ax.yaxis.grid(color="#E5E9EB", linestyle="-", linewidth=1)
                original_label = Patch(color="#3f93e1", label="original")
                synthetic_label = Patch(color="#ff9c54", label="synthetic")
                plt.axhline(y=0, color="#000000", linestyle="-")
                plt.legend(
                    handles=[original_label, synthetic_label],
                    loc="upper left",
                    bbox_to_anchor=(0, 1.09),
                    ncol=2,
                    frameon=False,
                )
                plt.savefig(
                    f"{self.reports_path}/utility_barplot.svg",
                    bbox_inches="tight",
                    format="svg",
                )
        log_msg = (
            "The ratio of synthetic {} accuracy to original is "
            "{}. The model considers the '{}' column "
            "as a target and other columns as predictors"
        )
        if best_binary is not None:
            score = (0 if score_binary == 0 else
                     round(synth_score_binary/score_binary, 4)
                     )
            logger.info(log_msg.format(
                'binary',
                score,
                best_binary
                )
            )
        if best_categ is not None:
            score = (0 if score_categ == 0 else
                     round(synth_score_categ/score_categ, 4)
                     )
            logger.info(log_msg.format(
                'multiclass',
                score,
                best_categ
                )
            )
        if best_regres is not None:
            score = (0 if score_regres == 0 else
                     abs(round(max(0, synth_regres_score) / score_regres, 4))
                     )
            logger.info(log_msg.format(
                'regression',
                score,
                best_regres
                )
            )
        logger.info("Utility metric is calculated")
        return result

    @staticmethod
    def __get_accuracy_score(y_true, y_pred, task_type):
        if task_type != "regression":
            score = accuracy_score(y_true=y_true, y_pred=y_pred)
        else:
            score = r2_score(y_true=y_true, y_pred=y_pred)
        return score

    def __model_process(self, model_object, targets, task_type):
        best_score = -1
        best_target = None
        best_model = None
        synthetic_score = -1

        for i, col in tqdm.tqdm(
            iterable=enumerate(targets),
            desc=f"Calculating utility metric for {task_type}",
            total=len(targets),
        ):
            if self.original.shape[1] < 2:
                logger.info(
                    "There is only one column in the dataset. "
                    "It will not be used in utility metric."
                )
                continue
            original = pd.get_dummies(self.original.drop(col, axis=1))
            original = StandardScaler().fit_transform(original)
            model_y = self.original[col].values

            # If the original dataset is large create a stratified sample
            if self.is_big_original_data_after_drop_na:
                original, _, model_y, _ = self.__perform_train_test_split(
                    original, model_y, self.sample_size
                )

            # validate number of unique values and instances in each class
            if not self._valid_target(col, model_y, task_type):
                continue

            (
                original_train,
                original_test,
                model_y_train,
                model_y_test
            ) = self.__perform_train_test_split(
                original,
                model_y,
                train_size=0.8
            )

            model = model_object.fit(
                X=original_train, y=model_y_train
            )
            score = self.__get_accuracy_score(
                model_y_test,
                model.predict(original_test),
                task_type,
            )
            if score > best_score:
                best_score = score
                best_target = col
                best_model = model

        # Calculate synthetic score
        if best_score > -1:
            if best_score < 0.6:
                logger.info(
                    f"The best score for all possible {task_type} models "
                    f"for the original data is "
                    f"{round(best_score, 4)}, which is below 0.6. "
                    f"The utility metric is unreliable"
                )
            synthetic = pd.get_dummies(self.synthetic.drop(best_target, axis=1))
            synthetic = StandardScaler().fit_transform(synthetic)
            synthetic_y = self.synthetic[best_target].values

            # If the synthetic dataset is large
            # create a stratified sample
            if self.is_big_synthetic_data_after_drop_na:
                synthetic, _, synthetic_y, _ = self.__perform_train_test_split(
                    synthetic, synthetic_y, self.sample_size
                )

            synthetic_score = self.__get_accuracy_score(
                synthetic_y,
                best_model.predict(synthetic),
                task_type,
            )

        return best_target, best_score, synthetic_score

    @staticmethod
    def __perform_train_test_split(data, model_y, train_size):
        """
        Splits the data into train and test sets.
        Tries stratified sampling first; if it fails,
        perform random sampling.
        """
        try:
            data_train, data_test, model_y_train, model_y_test = train_test_split(
                data,
                model_y,
                train_size=train_size,
                stratify=model_y,
                random_state=10
            )
        except ValueError:
            data_train, data_test, model_y_train, model_y_test = train_test_split(
                data,
                model_y,
                train_size=train_size,
                stratify=None,
                random_state=10
            )

        return data_train, data_test, model_y_train, model_y_test

    def __create_binary_class_models(self, binary_targets):
        best_target, score, synthetic_score = self.__model_process(
            LogisticRegression(n_jobs=-1, random_state=10),
            binary_targets, "binary classification"
        )
        return best_target, score, synthetic_score

    def __create_multi_class_models(self, multiclass_targets):
        best_target, score, synthetic_score = self.__model_process(
            RandomForestClassifier(n_jobs=-1, random_state=10),
            multiclass_targets,
            "multiclass classification",
        )
        return best_target, score, synthetic_score

    def __create_regression_models(self, cont_targets):
        best_target, score, synthetic_score = self.__model_process(
            RandomForestRegressor(n_jobs=-1, random_state=10),
            cont_targets, "regression"
        )
        return best_target, score, synthetic_score

    @staticmethod
    def _valid_target(col, model_y, task_type):
        """
        Validate the target column in utility metric calculation.

        Args:
        col (str): The name of the column being checked.
        model_y (array-like): The target values.
        task_type (str): type task

        Returns:
        bool: True if the column is valid
        """
        unique_values, unique_counts = np.unique(model_y, return_counts=True)

        # check if there is only one unique value in the column
        if len(unique_values) < 2:
            logger.info(
                f"Column '{col}' has only one unique value. "
                "It will not be used as target column "
                f"in utility metric calculation for {task_type}."
            )
            return False

        # check if there is more than 1 sample in each class
        if task_type in ["binary classification", "multiclass classification"]:
            if np.min(unique_counts) < 2:
                logger.info(
                    f"Column '{col}' has a class with less than 2 samples. "
                    "It will not be used as target column "
                    f"in utility metric calculation for {task_type}."
                )
                return False

        return True



import os
import shutil
import threading

import streamlit as st

from syngen.ml.data_loaders import DataLoader


UPLOAD_DIRECTORY = "uploaded_files"
MODEL_ARTIFACTS = "model_artifacts"


def show_data(uploaded_file):
    """
    Show sample data from the uploaded file
    """
    file_name = uploaded_file.name
    file_path = os.path.join(UPLOAD_DIRECTORY, file_name)
    if not os.path.exists(UPLOAD_DIRECTORY):
        os.makedirs(UPLOAD_DIRECTORY, exist_ok=True)
    with open(file_path, "wb") as file_object:
        file_object.write(uploaded_file.getvalue())
    df, schema = DataLoader(file_path).load_data()
    st.write(f"Preview of {file_name}:", df.head())
    st.write(f"Rows: {df.shape[0]}, columns: {df.shape[1]}")


def get_running_status():
    """
    Get the status of the process of a model training and generation data
    """
    if (
            ("gen_button" in st.session_state and st.session_state.gen_button is True) or
            [thread for thread in threading.enumerate() if thread.name == "train_and_infer"]
    ):
        return True
    else:
        return False


def set_session_state():
    """
    Set a session state for the Streamlit app
    """
    if "disabled" not in st.session_state:
        st.session_state.disabled = False


def cleanup_artifacts():
    """
    Clean up the artifacts
    """
    shutil.rmtree(UPLOAD_DIRECTORY, ignore_errors=True)
    shutil.rmtree(MODEL_ARTIFACTS, ignore_errors=True)



from typing import Literal
import os
from datetime import datetime
from queue import Queue
import sys
import traceback

from loguru import logger
from slugify import slugify
import streamlit as st
from streamlit.elements.widgets.file_uploader import UploadedFile
import streamlit.components.v1 as components

from syngen.ml.worker import Worker
from syngen.ml.utils import fetch_log_message, ProgressBarHandler


UPLOAD_DIRECTORY = "uploaded_files"
TIMESTAMP = slugify(datetime.now().strftime("%Y-%m-%d %H:%M:%S"))


class StreamlitHandler:
    """
    A class for handling the Streamlit app
    """
    def __init__(
            self,
            epochs: int,
            size_limit: int,
            reports: bool,
            uploaded_file: UploadedFile
    ):
        self.epochs = epochs
        self.size_limit = size_limit
        self.reports = ["accuracy"] if reports else []
        self.uploaded_file = uploaded_file
        self.file_name = self.uploaded_file.name
        self.table_name = os.path.splitext(self.file_name)[0]
        self.file_path = os.path.join(UPLOAD_DIRECTORY, self.file_name)
        self.sl_table_name = slugify(self.table_name)
        self.progress_handler = ProgressBarHandler()
        self.log_queue = Queue()
        self.log_error_queue = Queue()
        self.path_to_generated_data = (f"model_artifacts/tmp_store/{self.sl_table_name}/"
                                       f"merged_infer_{self.sl_table_name}.csv")
        self.path_to_report = (f"model_artifacts/tmp_store/{self.sl_table_name}/"
                               f"reports/accuracy_report.html")
        self.train_settings = {
            "source": self.file_path,
            "epochs": self.epochs,
            "row_limit": 10000,
            "drop_null": False,
            "batch_size": 32,
            "reports": []
        }
        self.infer_settings = {
            "size": self.size_limit,
            "batch_size": 32,
            "run_parallel": False,
            "random_seed": None,
            "reports": self.reports,
        }

    def set_logger(self):
        """
        Set a logger to see logs, and collect log messages
        with the log level - 'INFO' in a log file and stdout
        """
        logger.remove()
        logger.add(self.console_sink, colorize=True, level="INFO")
        logger.add(self.file_sink, level="INFO")

    def console_sink(self, message: str):
        """
        Put log messages to a log queue
        """
        log_message = fetch_log_message(message)
        sys.stdout.write(log_message + "\n")
        self.log_queue.put(log_message)

    def file_sink(self, message: str):
        """
        Write log messages to a log file
        """
        path_to_logs = f"model_artifacts/system_store/{self.sl_table_name}_{TIMESTAMP}.log"
        os.environ["SUCCESS_LOG_FILE"] = path_to_logs
        os.makedirs(os.path.dirname(path_to_logs), exist_ok=True)
        with open(path_to_logs, "a") as log_file:
            log_message = fetch_log_message(message)
            log_file.write(log_message + "\n")

    def _get_worker(self, process_type: Literal["train", "infer"]):
        """
        Get a Worker object

        Parameters
        ----------
        process_type : str
            The type of process ("train" or "infer").

        Returns
        -------
        Worker
            The Worker object.
        """
        worker = Worker(
            table_name=self.table_name,
            settings=self.train_settings if process_type == "train" else self.infer_settings,
            metadata_path=None,
            log_level="INFO",
            type_of_process=process_type
        )
        return worker

    def train_model(self):
        """
        Launch a model training
        """
        try:
            logger.info("Starting model training...")
            worker = self._get_worker("train")
            ProgressBarHandler().set_progress(0.01)
            worker.launch_train()
            logger.info("Model training completed")
        except Exception as e:
            error_message = (f"The error raised during the training process - "
                             f"{traceback.format_exc()}")
            logger.error(error_message)
            self.log_error_queue.put(e)
            raise e

    def infer_model(self):
        """
        Launch a data generation
        """
        try:
            logger.info("Starting data generation...")
            worker = self._get_worker("infer")
            worker.launch_infer()
            logger.info("Data generation completed")
        except Exception as e:
            error_message = (f"The error raised during the inference process - "
                             f"{traceback.format_exc()}")
            logger.error(error_message)
            self.log_error_queue.put(e)
            raise e

    def set_monitoring(self):
        """
        Reset the progress bar and set up the logger
        """
        self.progress_handler.reset_instance()
        self.set_logger()

    def train_and_infer(self):
        """
        Launch a model training and data generation
        """
        self.set_monitoring()
        self.train_model()
        self.infer_model()

    @staticmethod
    def generate_button(label: str, path_to_file: str, download_name: str):
        """
        Generate a download button
        """
        if os.path.exists(path_to_file):
            with open(path_to_file, "rb") as f:
                st.download_button(
                    label,
                    f,
                    file_name=download_name,
                )

    def open_report(self):
        """
        Open the accuracy report in the iframe
        """
        if os.path.exists(self.path_to_report) and self.reports:
            with open(self.path_to_report, "r") as report:
                report_content = report.read()
            with st.expander("View the accuracy report"):
                components.html(report_content, 680, 1000, True)

    def generate_buttons(self):
        """
        Generate download buttons for downloading artifacts
        """
        self.generate_button(
            "Download generated data",
            self.path_to_generated_data,
            f"generated_data_{self.sl_table_name}.csv"
        )
        self.generate_button(
            "Download logs",
            os.getenv("SUCCESS_LOG_FILE", ""),
            f"logs_{self.sl_table_name}.log"
        )
        if self.reports:
            self.generate_button(
                "Download the accuracy report",
                self.path_to_report,
                f"accuracy_report_{self.sl_table_name}.html"
            )
            self.open_report()



from typing import Tuple, Optional, Dict, List, Callable
from abc import ABC, abstractmethod
import os
import math
from ulid import ULID
from uuid import UUID
from copy import deepcopy

import pandas as pd
import numpy as np
from numpy.random import seed, choice
from pathos.multiprocessing import ProcessingPool
import dill
from scipy.stats import gaussian_kde
from collections import OrderedDict
from tensorflow.keras.preprocessing.text import Tokenizer
from slugify import slugify
from loguru import logger
from attrs import define, field

from syngen.ml.vae import *  # noqa: F403
from syngen.ml.data_loaders import DataLoader
from syngen.ml.reporters import Report
from syngen.ml.vae.models.dataset import Dataset
from syngen.ml.utils import (
    fetch_config,
    check_if_features_assigned,
    get_initial_table_name,
    ProgressBarHandler
)
from syngen.ml.context import get_context


class AbstractHandler(ABC):
    @abstractmethod
    def set_next(self, handler):
        pass

    @abstractmethod
    def handle(self, data: pd.DataFrame, **kwargs):
        pass


@define
class BaseHandler(AbstractHandler):
    metadata: Dict = field(kw_only=True)
    paths: Dict = field(kw_only=True)
    table_name: str = field(kw_only=True)
    loader: Optional[Callable[[str], pd.DataFrame]] = None
    _next_handler: Optional[AbstractHandler] = None

    def set_next(self, handler: AbstractHandler) -> AbstractHandler:
        self._next_handler = handler
        return handler

    @abstractmethod
    def handle(self, data: Optional[pd.DataFrame], **kwargs):
        if self._next_handler:
            return self._next_handler.handle(data, **kwargs)

        return None

    @staticmethod
    def create_wrapper(
        cls_name, data: Optional[pd.DataFrame] = None, schema: Optional[Dict] = None, **kwargs
    ):
        return globals()[cls_name](
            data,
            schema,
            metadata=kwargs["metadata"],
            table_name=kwargs["table_name"],
            paths=kwargs["paths"],
            batch_size=kwargs["batch_size"],
            main_process=kwargs["main_process"],
            process=kwargs["process"],
        )


@define
class RootHandler(BaseHandler):

    def handle(self, data: pd.DataFrame, **kwargs):
        return super().handle(data, **kwargs)


@define
class LongTextsHandler(BaseHandler):
    schema: Optional[Dict] = field(kw_only=True)

    @staticmethod
    def series_count_words(x):
        return len(str(x).split())

    def _prepare_dir(self):
        os.makedirs(self.paths["no_ml_state_path"], exist_ok=True)

    def _save_no_ml_checkpoints(self, features: Dict):
        with open(f'{self.paths["no_ml_state_path"]}kde_params.pkl', "wb") as file:
            dill.dump(features, file)

    def handle(self, data: pd.DataFrame, **kwargs):
        self._prepare_dir()

        dataset = fetch_config(self.paths["dataset_pickle_path"])
        long_text_columns = dataset.long_text_columns

        if len(long_text_columns) > 0:
            features = {}
            for col in long_text_columns:
                tokenizer = Tokenizer(lower=False, char_level=True)
                if type(data[col].dropna().values[0]) is bytes:
                    text_col = data[col].str.decode("utf-8", errors="ignore")
                else:
                    text_col = data[col]
                text_col = text_col.fillna("")
                tokenizer.fit_on_texts(text_col)

                indexes = OrderedDict((k, v) for k, v in tokenizer.word_index.items() if k != " ")
                counts = OrderedDict((k, v) for k, v in tokenizer.word_counts.items() if k != " ")
                ordered_indexes = OrderedDict((k, indexes[k]) for k in counts.keys())
                text_structure = np.array(
                    [text_col.str.len(), text_col.apply(self.series_count_words)]
                )
                noise_to_prevent_singularity = np.random.uniform(
                    low=-1e-4,
                    high=1e-4,
                    size=(text_structure.shape[0], text_structure.shape[1]),
                )
                bw_width = text_structure.shape[1] / text_structure.shape[1] ** 1.3
                kde = gaussian_kde(
                    (text_structure + noise_to_prevent_singularity).astype("float64"),
                    bw_method=bw_width,
                )
                features[col] = {
                    "counts": counts,
                    "indexes": ordered_indexes,
                    "kde": kde,
                }

            self._save_no_ml_checkpoints(features)

        else:
            logger.info("No columns to train kde over found")
        return super().handle(data, **kwargs)


@define
class VaeTrainHandler(BaseHandler):
    wrapper_name: str = field(kw_only=True)
    schema: Dict = field(kw_only=True)
    epochs: int = field(kw_only=True)
    row_subset: int = field(kw_only=True)
    drop_null: bool = field(kw_only=True)
    batch_size: int = field(kw_only=True)
    type_of_process: str = field(kw_only=True)
    reports: List[str] = field(kw_only=True)

    def __fit_model(self, data: Optional[pd.DataFrame]):
        logger.info("Start VAE training")
        if data is None:
            logger.error("For mode = 'train' path must be provided")
            raise ValueError("Can't read data from path or path is None")

        self.model = self.create_wrapper(
            self.wrapper_name,
            data,
            self.schema,
            metadata=self.metadata,
            table_name=self.table_name,
            paths=self.paths,
            batch_size=self.batch_size,
            main_process=self.type_of_process,
            process="train",
        )
        self.model.batch_size = min(self.batch_size, len(data))
        list_of_reports = [f'\'{report}\'' for report in self.reports]
        list_of_reports = ', '.join(list_of_reports) if list_of_reports else '\'none\''
        logger.debug(
            f"Train model with parameters: epochs={self.epochs}, "
            f"row_subset={self.row_subset}, drop_null={self.drop_null}, "
            f"batch_size={self.batch_size}, reports - {list_of_reports}"
        )

        self.model.fit_on_df(epochs=self.epochs)

        if not check_if_features_assigned(self.paths["dataset_pickle_path"]):
            return

        self.model.save_state(self.paths["state_path"])
        log_message = "Finished VAE training"
        logger.info(log_message)
        ProgressBarHandler().set_progress(message=log_message)

    def __prepare_dir(self):
        os.makedirs(self.paths["fk_kde_path"], exist_ok=True)

    def handle(self, data: pd.DataFrame, **kwargs):
        self.__prepare_dir()
        self.__fit_model(data)
        return super().handle(data, **kwargs)


@define
class VaeInferHandler(BaseHandler):
    metadata_path: str = field(kw_only=True)
    random_seed: Optional[int] = field(kw_only=True)
    size: int = field(kw_only=True)
    batch_size: int = field(kw_only=True)
    run_parallel: bool = field(kw_only=True)
    reports: List[str] = field(kw_only=True)
    wrapper_name: str = field(kw_only=True)
    log_level: str = field(kw_only=True)
    type_of_process: str = field(kw_only=True)
    random_seed_list: List = field(init=False)
    vae: Optional[VAEWrapper] = field(init=False)  # noqa: F405
    dataset: Dataset = field(init=False)
    original_schema: Dict = field(init=False)
    has_vae: bool = field(init=False)
    has_no_ml: bool = field(init=False)
    batch_num: int = field(init=False)

    def __attrs_post_init__(self):
        if self.random_seed:
            seed(self.random_seed)
        self.batch_num = math.ceil(self.size / self.batch_size)
        self.random_seeds_list = list()
        self.vae = None
        self.dataset = fetch_config(self.paths["dataset_pickle_path"])
        path_to_schema = self.paths["original_schema_path"]
        self.original_schema = (
            fetch_config(path_to_schema) if os.path.exists(path_to_schema) else None
        )
        self.has_vae = len(self.dataset.features) > 0

        if self.has_vae:
            self.vae = self.__get_wrapper()

        self.has_no_ml = os.path.exists(f'{self.paths["path_to_no_ml"]}')

    @staticmethod
    def _synth_word(size, indexes, counts):
        generator = np.random.default_rng()
        return (
            "".join(
                generator.choice(
                    np.array(list(indexes)),
                    size=size,
                    p=np.array(list(counts.values())) / sum(np.array(list(counts.values())))
                )
            )
        )

    def __get_wrapper(self):
        """
        Create and get the wrapper for the VAE model
        """
        return self.create_wrapper(
            self.wrapper_name,
            metadata=deepcopy(self.metadata),
            table_name=self.table_name,
            paths=self.paths,
            batch_size=self.batch_size,
            main_process=self.type_of_process,
            process="infer",
        )

    def _prepare_dir(self):
        tmp_store_path = self.paths["tmp_store_path"]
        os.makedirs(tmp_store_path, exist_ok=True)

    def _is_pk(self):
        is_pk = self.table_name.endswith("_pk")
        return is_pk

    def _concat_slices_with_unique_pk(self, df_slices: list):
        if self.metadata and self.table_name in self.metadata:
            config_of_keys = self.metadata.get(self.table_name).get("keys", {})
            for key in config_of_keys.keys():
                column = config_of_keys.get(key).get("columns")[0]
                if config_of_keys.get(key).get("type") == "PK" and not isinstance(
                    df_slices[0][column][0], (str, UUID, ULID)
                ):
                    cumm_len = 0
                    for i, frame in enumerate(df_slices):
                        frame[column] = frame[column].map(lambda pk_val: pk_val + cumm_len)
                        cumm_len += len(frame)
        return pd.concat(df_slices, ignore_index=True)

    def generate_vae(self, size):
        synthetic_infer = self.vae.predict_sampled_df(size)
        return synthetic_infer

    def generate_long_texts(self, size, synthetic_infer):
        with open(f'{self.paths["path_to_no_ml"]}', "rb") as file:
            features = dill.load(file)
        for col in features.keys():
            kde = features[col]["kde"]
            text_structures = np.maximum(kde.resample(size).astype("int32"), 0)
            indexes = features[col]["indexes"]
            counts = features[col]["counts"]
            generated_column = [
                " ".join(
                    [
                        self._synth_word(s, indexes, counts)
                        for s in np.maximum(np.random.normal(i / j, 1, j).astype("int32"), 2)
                    ]
                )
                for i, j in zip(*text_structures)
            ]
            logger.debug(f"Long text for column '{col}' is generated.")
            synthetic_infer[col] = generated_column
        return synthetic_infer

    def run_separate(self, params: Tuple):
        i, size = params

        if self.random_seed:
            seed(self.random_seeds_list[i])

        synthetic_infer = pd.DataFrame()

        if self.has_vae:
            logger.info(f"VAE generation for '{self.table_name}' started.")
            synthetic_infer = self.generate_vae(size)
        if self.has_no_ml:
            logger.info(f"Long texts generation for '{self.table_name}' started.")
            synthetic_infer = self.generate_long_texts(size, synthetic_infer)

        return synthetic_infer

    def split_by_batches(self):
        """
        Split the total size into batches
        based on the specified number of nodes.

        This method divides the total size
        into a specified number of batches (nodes).
        Each batch will have a size equal to batch_size,
        except for the last batch, which will contain the remaining size
        """
        quote = self.batch_size
        nodes = self.batch_num
        data = [quote] * (nodes - 1)
        data.append(self.size - quote * (nodes - 1))
        return data

    def run(self, size: int, run_parallel: bool):
        logger.info("Start data synthesis")
        if run_parallel:
            pool = ProcessingPool()
            if self.random_seed:
                self.random_seeds_list = choice(
                    range(0, max(100, pool.nodes)), pool.nodes, replace=False
                )

            frames = pool.map(
                self.run_separate, enumerate(self.split_by_batches())
            )
            generated = self._concat_slices_with_unique_pk(frames)
        else:
            if self.random_seed:
                self.random_seeds_list = [self.random_seed]
            generated = self.run_separate((0, size))
        return generated

    def kde_gen(self, pk_table, pk_column_label, size, fk_label):
        pk = pk_table[pk_column_label]

        try:
            with open(f'{self.paths["fk_kde_path"]}{fk_label}.pkl', "rb") as file:
                kde = dill.load(file)
            pk = pk.dropna()
            numeric_pk = np.arange(len(pk)) if pk.dtype == "object" else pk
            fk_pdf = np.maximum(kde.evaluate(numeric_pk), 1e-12)
            synth_fk = np.random.choice(pk, size=size, p=fk_pdf / sum(fk_pdf), replace=True)
            synth_fk = pd.DataFrame({fk_label: synth_fk}).reset_index(drop=True)

        except FileNotFoundError:
            logger.warning(
                f"The mapper for the '{fk_label}' text key is not found. Making simple sampling"
            )
            synth_fk = pk.sample(size, replace=True).reset_index(drop=True)
            synth_fk = synth_fk.rename(fk_label)

        return synth_fk

    def _get_pk_path(self, pk_table, table_name) -> str:
        """
        Set the path to synthetic data of corresponding pk table
        """
        destination_to_pk_table = None
        if self.type_of_process == "infer":
            infer_settings = self.metadata[pk_table].get("infer_settings", {})
            destination_to_pk_table = infer_settings.get("destination")

            if destination_to_pk_table is None:
                destination_to_pk_table = (
                    f"model_artifacts/tmp_store/{slugify(pk_table)}/"
                    f"merged_infer_{slugify(pk_table)}.csv"
                )
        initial_table_name = get_initial_table_name(table_name)
        if self.type_of_process == "train":
            destination_to_pk_table = self.paths["path_to_merged_infer"].replace(
                slugify(initial_table_name), slugify(pk_table)
            )
        if not os.path.exists(destination_to_pk_table):
            raise FileNotFoundError(
                "The table with a primary key specified in the metadata file does not "
                "exist or is not trained. Ensure that the metadata contains the "
                "name of referenced table with a primary key in the foreign key "
                "declaration section."
            )

        return destination_to_pk_table

    def generate_keys(self, generated, size, metadata, table_name):
        metadata_of_table = metadata.get(table_name)
        if "keys" not in metadata_of_table:
            return None
        config_of_keys = metadata_of_table.get("keys")
        for key in config_of_keys.keys():
            if config_of_keys.get(key).get("type") == "FK":
                fk_column_name = config_of_keys.get(key).get("columns")[0]
                pk_table = config_of_keys.get(key).get("references").get("table")
                pk_path = self._get_pk_path(pk_table=pk_table, table_name=table_name)
                pk_table_data, pk_table_schema = DataLoader(path=pk_path).load_data()
                pk_column_label = config_of_keys.get(key).get("references").get("columns")[0]
                logger.info(f"The '{pk_column_label}' assigned as a foreign_key feature")

                synth_fk = self.kde_gen(pk_table_data, pk_column_label, size, fk_column_name)
                generated = generated.reset_index(drop=True)

                null_column_name = f"{key}_null"
                if null_column_name in generated.columns:
                    not_null_column_mask = generated[null_column_name].astype("float64") <= 0.5
                    synth_fk = synth_fk.where(not_null_column_mask, np.nan)
                    generated = generated.drop(null_column_name, axis=1)

                generated[fk_column_name] = synth_fk
        return generated

    def _restore_empty_columns(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Restore empty columns in the generated table
        """
        empty_columns = self.dataset.dropped_columns

        empty_df = pd.DataFrame(index=df.index, columns=list(empty_columns))
        df = pd.concat([df, empty_df], axis=1)

        return df

    def _drop_technical_columns(self, prepared_data: pd.DataFrame) -> pd.DataFrame:
        """
        Workaround for the case when all columns are dropped with technical column
        """
        tech_columns = list(self.dataset.tech_columns)
        if tech_columns:
            prepared_data = prepared_data.drop(tech_columns, axis=1)
            logger.debug(
                f"Technical columns {tech_columns} were removed from the generated table."
            )
            Report().unregister_reporters(self.table_name)
            logger.info(
                "Since there were no columns suitable for training, reports will not be generated "
                f"for the table '{self.table_name}'."
            )
        return prepared_data

    def _save_data(self, generated_data):
        """
        Save generated data to the path
        """
        DataLoader(path=self.paths["path_to_merged_infer"]).save_data(
            data=generated_data,
            format=get_context().get_config(),
        )

    def handle(self, **kwargs):
        self._prepare_dir()
        list_of_reports = [f'\'{report}\'' for report in self.reports]
        list_of_reports = ', '.join(list_of_reports) if list_of_reports else '\'none\''
        log_message = (
            f"Infer model with parameters: size={self.size}, "
            f"run_parallel={self.run_parallel}, batch_size={self.batch_size}, "
            f"random_seed={self.random_seed}"
        )
        if self.type_of_process == "infer":
            log_message += f", reports - {list_of_reports}"
        logger.debug(log_message)
        logger.info(f"Total of {self.batch_num} batch(es)")
        batches = self.split_by_batches()
        delta = ProgressBarHandler().delta / self.batch_num
        prepared_batches = []
        for i, batch in enumerate(batches):
            log_message = (f"Data synthesis for the table - '{self.table_name}'. "
                           f"Generating the batch {i + 1} of {self.batch_num}")
            ProgressBarHandler().set_progress(
                progress=ProgressBarHandler().progress + delta,
                delta=delta,
                message=log_message,
            )
            logger.info(log_message)
            prepared_batch = self.run(batch, self.run_parallel)
            prepared_batches.append(prepared_batch)
        prepared_data = (
            self._concat_slices_with_unique_pk(prepared_batches)
            if len(prepared_batches) > 0
            else pd.DataFrame()
        )
        prepared_data = self._restore_empty_columns(prepared_data)

        prepared_data = self._drop_technical_columns(prepared_data)

        is_pk = self._is_pk()

        if self.metadata_path is not None:
            if not is_pk:
                generated_data = self.generate_keys(
                    prepared_data, self.size, self.metadata, self.table_name
                )
                generated_data = generated_data[self.dataset.order_of_columns]

                if generated_data is None:
                    self._save_data(prepared_data)
                else:
                    self._save_data(generated_data)
            else:
                self._save_data(prepared_data)
        if self.metadata_path is None:
            prepared_data = prepared_data[self.dataset.order_of_columns]

            self._save_data(prepared_data)



from typing import Optional, Dict, Any, Union
import os
import requests
from statistics import mean, StatisticsError

import mlflow
from loguru import logger
from syngen.ml.utils import fetch_unique_root


HARDWARE_METRICS = [
    "system/cpu_utilization_percentage",
    "system/system_memory_usage_percentage",
    "system/system_memory_usage_megabytes",
    "system/gpu_utilization_percentage",
    "system/gpu_memory_usage_percentage",
    "system/gpu_memory_usage_megabytes"
]


class MlflowTrackerFactory:
    """
    A factory class for creating the Mlflow tracker
    """

    @staticmethod
    def check_mlflow_server(server_url):
        """
        Check if the MlFlow server is up and running
        """
        if server_url is None:
            logger.warning("MLFlow server URL not provided")
            return False
        try:
            response = requests.get(server_url)
            # If the response was successful, no Exception will be raised
            response.raise_for_status()
        except requests.exceptions.HTTPError as http_err:
            logger.warning(
                f"An HTTP error occurred while connecting to the MLFlow server: {http_err}"
            )
        except Exception as err:
            logger.warning(
                f"An unexpected error occurred while connecting to the MLFlow server: {err}"
            )
        else:
            logger.info("MLFlow server is up and running")
            return True

    @classmethod
    def create_tracker(
        cls, table_name: Optional[str], metadata_path: Optional[str], is_active=False
    ):
        """
        Create the Mlflow tracker, and create or set the experiment
        """
        experiment_name = cls.get_mlflow_exp_name(table_name, metadata_path)

        tracker = MlflowTracker(experiment_name, is_active)

        response = cls.check_mlflow_server(os.environ.get("MLFLOW_TRACKING_URI"))

        if response:
            tracker.is_active = True
            tracker.connect_to_server = True
        else:
            tracker.is_active = False
            tracker.connect_to_server = False
            logger.warning(
                "MLFlow server is either unreachable or not set up, "
                "therefore the tracking will not be performed"
            )

        tracker.set_tracking_uri(os.environ.get("MLFLOW_TRACKING_URI"))
        tracker.set_experiment(experiment_name)

    @classmethod
    def get_mlflow_exp_name(
        cls, table_name: Optional[str], metadata_path: Optional[str]
    ) -> str:
        """
        Get the name of the Mlflow experiment
        """
        if os.getenv("MLFLOW_EXPERIMENT_NAME"):
            return os.getenv("MLFLOW_EXPERIMENT_NAME")
        return fetch_unique_root(table_name, metadata_path)


class MlflowTracker:
    """
    A singleton class for tracking the Mlflow experiments.
    All methods are derived from the Mlflow API,
    however a check for the active state is added.
    """

    _instance = None

    def __new__(cls, experiment_name=None, is_active=False):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.experiment_name = experiment_name
            cls._instance.connect_to_server = is_active
            cls._instance.is_active = is_active
        return cls._instance

    def log_metric(self, key: str, value: float, step: Optional[int] = None):
        """
        Log a metric to the current run
        """
        if self.is_active:
            mlflow.log_metric(key, value, step)

    def log_duration(self, run_id, table, stage):
        """
        Log the duration of the run
        """
        try:
            run_info = self.get_run(run_id).info
            duration = (run_info.end_time - run_info.start_time) / 1000
            self.log_metric(key=f'{table}-{stage}-duration', value=duration)
        except AttributeError:
            pass

    def log_hardware_metrics(self, run_id, table, stage):
        """
        Log the hardware metrics of the run
        """
        for metric in HARDWARE_METRICS:
            try:
                mean_value = mean([m.value for m in self.get_metric_history(run_id, metric)])
                self.log_metric(key=f'{table}-{stage}-{metric}', value=mean_value)
            except (TypeError, StatisticsError):
                pass

    def collect_metrics(self, table, stage):
        """
        Collect the duration of the run
        and hardware metrics related to the run
        """
        runs = self.search_runs(table_name=table, type_of_process=stage)
        for run_id in runs:
            self.log_duration(run_id, table, stage)
            self.log_hardware_metrics(run_id, table, stage)

    def log_artifact(self, local_path: str, artifact_path: Optional[str] = None):
        """
        Log an artifact from the local path to the current run
        """
        if self.is_active:
            mlflow.log_artifact(local_path, artifact_path)

    def log_params(self, params: Dict[str, Any]):
        if self.is_active:
            mlflow.log_params(params)

    def start_run(
        self,
        run_id: str = None,
        experiment_id: Optional[str] = None,
        run_name: Optional[str] = None,
        nested: bool = False,
        tags: Optional[Dict[str, Any]] = None,
        description: Optional[str] = None,
    ):
        if self.is_active:
            mlflow.start_run(
                run_id=run_id,
                experiment_id=experiment_id,
                run_name=run_name,
                nested=nested,
                tags=tags,
                description=description
            )

    def get_last_run(self):
        if self.is_active:
            return mlflow.last_active_run()

    def end_run(self):
        if self.is_active:
            mlflow.end_run()

    def set_tracking_uri(self, uri: str):
        """
        Set the tracking URI
        """
        if self.is_active:
            mlflow.set_tracking_uri(uri)

    def create_experiment(self, name: str, artifact_location: Optional[str] = None):
        if self.is_active:
            mlflow.create_experiment(name, artifact_location)

    @staticmethod
    def get_experiment(experiment_name: str):
        """
        Get the last matching of the experiment by name
        """
        experiments = mlflow.search_experiments(
            filter_string=f"name LIKE '{experiment_name}'"
        )
        return experiments[0] if experiments else []

    def set_experiment(self, experiment_name: str = None, experiment_id: str = None):
        """
        Set the experiment for tracking.
        If the experiment name is not provided, the last experiment will be used.
        """
        if self.is_active:
            last_matching = self.get_experiment(experiment_name)
            env_value = os.getenv("MLFLOW_EXPERIMENT_NAME", "")
            if not last_matching:
                MlflowTracker().create_experiment(
                    experiment_name,
                    artifact_location=os.environ.get(
                        "MLFLOW_ARTIFACTS_DESTINATION", "/mlflow_tracker"
                    ),
                )
                mlflow.set_experiment(experiment_name, experiment_id)

                if not env_value:
                    logger.warning(
                        f"A new experiment with the name based on 'table_name' or "
                        f"'metadata_path' value - '{experiment_name}' will be created"
                    )

                if env_value:
                    logger.info(
                        f"A new experiment with the name - '{experiment_name}' will be created"
                    )

            if last_matching:
                matching_name = last_matching.name
                mlflow.set_experiment(matching_name, experiment_id)
                if env_value:
                    logger.warning(
                        f"The experiment with the same name - '{experiment_name}' already exists. "
                        f"The created runs will be stored in the experiment - '{matching_name}'"
                    )
                if not env_value:
                    logger.warning(
                        f"The experiment with the name similar to 'table_name' or "
                        f"'metadata_path' value - '{experiment_name}' already exists "
                        f"and new runs will be sent there"
                    )

    def log_metrics(
        self, metrics: Dict[str, Union[float, str]], step: Optional[int] = None
    ):
        if self.is_active:
            mlflow.log_metrics(metrics, step)

    def search_runs(self, table_name: str, type_of_process: str):
        """
        Get the list of runs related the certain experiment
        """
        if self.is_active:
            run = mlflow.search_runs(
                experiment_names=[self.experiment_name],
                filter_string=f"run_name like '{table_name}-{type_of_process}%'"
            )
            try:
                count_runs = 2 if run.at[0, "tags.mlflow.runName"].endswith("-2") else 1
                return run["run_id"][:count_runs].to_list()
            except KeyError:
                return []
        else:
            return []

    def get_run(self, run_id):
        """
        Get the run by the run_id
        """
        if self.is_active:
            return mlflow.get_run(run_id)

    def get_metric_history(self, run_id, metric_key):
        if self.is_active:
            return mlflow.tracking.MlflowClient().get_metric_history(run_id, metric_key)



from typing import Dict
import copy


class Context:
    def __init__(self):
        self.config = {}

    def set_config(self, value: Dict):
        self.config = copy.deepcopy(value)

    def get_config(self) -> Dict:
        return self.config


# Singleton pattern to ensure there is only one instance of Configuration
_config_instance: Context = None


def get_context() -> Context:
    global _config_instance
    if _config_instance is None:
        _config_instance = Context()
    return _config_instance


def global_context(metadata: Dict):
    global_config = get_context()
    global_config.set_config(metadata)



from abc import ABC, abstractmethod
import os
import traceback

from loguru import logger
from copy import deepcopy

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"

from syngen.ml.handlers import RootHandler
from syngen.ml.reporters import Report, AccuracyReporter, SampleAccuracyReporter
from syngen.ml.config import TrainConfig, InferConfig
from syngen.ml.handlers import LongTextsHandler, VaeTrainHandler, VaeInferHandler
from syngen.ml.vae import VanillaVAEWrapper
from syngen.ml.data_loaders import BinaryLoader
from syngen.ml.mlflow_tracker.mlflow_tracker import MlflowTracker
from syngen.ml.utils import get_initial_table_name, clean_up_metadata


class Strategy(ABC):
    """
    Abstract class for the strategies of training or infer process
    """

    def __init__(self):
        self.handler = None
        self.config = None
        self.metadata = None

    @abstractmethod
    def run(self, *args, **kwargs):
        pass

    @abstractmethod
    def set_config(self):
        pass

    @abstractmethod
    def add_handler(self, *args, **kwargs):
        pass

    @abstractmethod
    def add_reporters(self):
        """
        Set up reporter which used in order to create the sampling report during training process
        """
        pass


class TrainStrategy(Strategy, ABC):
    """
    Class of a strategy defined in a training process
    """

    def _save_training_config(self):
        metadata = deepcopy(self.config.metadata)
        self.config.metadata = clean_up_metadata(metadata=metadata)

        BinaryLoader(
            path=self.config.paths["train_config_pickle_path"]
        ).save_data(data=self.config)

    def set_config(self, **kwargs):
        """
        Set up configuration for training process
        """
        configuration = TrainConfig(**kwargs)
        self.config = configuration
        self.metadata = deepcopy(self.config.metadata)
        self._save_training_config()
        return self

    def add_handler(self):
        """
        Set up the handler which used in training process
        """
        root_handler = RootHandler(
            metadata=self.metadata,
            table_name=self.config.table_name,
            paths=self.config.paths,
            loader=self.config.loader
        )

        vae_handler = VaeTrainHandler(
            metadata=self.metadata,
            table_name=self.config.table_name,
            schema=self.config.schema,
            paths=self.config.paths,
            wrapper_name=VanillaVAEWrapper.__name__,
            epochs=self.config.epochs,
            row_subset=self.config.row_subset,
            drop_null=self.config.drop_null,
            batch_size=self.config.batch_size,
            reports=self.config.reports,
            type_of_process="train",
        )

        long_text_handler = LongTextsHandler(
            metadata=self.metadata,
            table_name=self.config.table_name,
            schema=self.config.schema,
            paths=self.config.paths,
        )

        root_handler.set_next(vae_handler).set_next(long_text_handler)

        self.handler = root_handler
        return self

    def add_reporters(self, **kwargs):
        # TODO: now the reporter isn't added if the flatten metadata exists
        # This should be refactored in the future
        table_name = self.config.table_name
        flatten_metadata_exists = os.path.exists(self.config.paths["path_to_flatten_metadata"])
        if flatten_metadata_exists:
            logger.warning(
                "The sample report isn't available for a table containing JSON column(s)"
            )
        if (
                not table_name.endswith("_fk")
                and "sample" in self.config.reports
                and not flatten_metadata_exists
        ):
            sample_reporter = SampleAccuracyReporter(
                table_name=get_initial_table_name(table_name),
                paths=self.config.paths,
                config=self.config.to_dict(),
                metadata=self.metadata,
            )
            Report().register_reporter(table=table_name, reporter=sample_reporter)

        return self

    def run(self, **kwargs):
        """
        Launch the training process
        """
        try:
            table = kwargs["table_name"]
            data = kwargs["data"]
            # Start the separate run for the preprocess stage
            # included preprocessing of the original data, identification of data types of columns,
            # fit and transform of the assigned features
            MlflowTracker().start_run(
                run_name=f"{table}-PREPROCESS",
                tags={"table_name": table, "process": "preprocess"},
            )
            self.set_config(**kwargs)
            self.add_reporters().add_handler()
            self.handler.handle(data)
            # End the separate run for the training stage
            MlflowTracker().end_run()
        except Exception:
            logger.error(
                f"Training of the table - \'{kwargs['table_name']}\' failed on running stage.\n"
                f"The traceback of the error - {traceback.format_exc()}"
            )
            raise
        else:
            logger.info(f"Training of the table - \'{kwargs['table_name']}\' was completed")


class InferStrategy(Strategy):
    """
    Class of a strategy defined in the infer process
    """

    def set_config(self, **kwargs):
        """
        Set up the configuration for infer process
        """
        configuration = InferConfig(**kwargs)
        self.config = configuration
        self.metadata = deepcopy(self.config.metadata)
        return self

    def add_handler(self, type_of_process: str):
        """
        Set up the handler which used in infer process
        """
        self.handler = VaeInferHandler(
            metadata_path=self.config.metadata_path,
            metadata=self.metadata,
            table_name=self.config.table_name,
            paths=self.config.paths,
            wrapper_name=VanillaVAEWrapper.__name__,
            size=self.config.size,
            random_seed=self.config.random_seed,
            batch_size=self.config.batch_size,
            run_parallel=self.config.run_parallel,
            reports=self.config.reports,
            log_level=self.config.log_level,
            type_of_process=type_of_process,
            loader=self.config.loader
        )
        return self

    def add_reporters(self):
        table_name = self.config.table_name
        if (
                not table_name.endswith("_fk") and
                any([item in ["accuracy", "metrics_only"] for item in self.config.reports])
        ):
            accuracy_reporter = AccuracyReporter(
                table_name=get_initial_table_name(table_name),
                paths=self.config.paths,
                config=self.config.to_dict(),
                metadata=self.metadata,
                loader=self.config.loader,
            )
            Report().register_reporter(table=table_name, reporter=accuracy_reporter)

        return self

    def run(self, **kwargs):
        """
        Launch the infer process
        """
        table_name = kwargs["table_name"]
        try:
            self.set_config(**kwargs)
            MlflowTracker().log_params(self.config.to_dict())
            self.add_reporters()
            self.add_handler(type_of_process=kwargs["type_of_process"])
            self.handler.handle()
        except Exception:
            logger.error(
                f"Generation of the table - '{table_name}' failed on running stage.\n"
                f"The traceback of the error - {traceback.format_exc()}"
            )
            raise
        else:
            logger.info(
                f"Synthesis of the table - '{table_name}' was completed. "
                f"Synthetic data saved in \'{self.handler.paths['path_to_merged_infer']}\'"
            )



from pathlib import Path

import tensorflow as tf
import pickle
from tensorflow.keras.models import Model
from tensorflow.keras.layers import (
    Input,
    Dense,
    Dropout,
    LeakyReLU,
    concatenate,
    Lambda,
    BatchNormalization,
    Activation,
)
from sklearn.mixture import BayesianGaussianMixture
import numpy as np
import pandas as pd
from loguru import logger

from syngen.ml.vae.models.custom_layers import FeatureLossLayer
from syngen.ml.utils import slugify_parameters, ProgressBarHandler


class CVAE:
    """
    A class implementing the model architecture.
    """

    def __init__(self, dataset, batch_size, latent_dim, intermediate_dim, latent_components):
        self.dataset = dataset
        self.intermediate_dim = intermediate_dim
        self.batch_size = batch_size
        self.latent_dim = latent_dim
        self.latent_components = min(latent_components, len(self.dataset.order_of_columns))
        self.model = None
        self.latent_model = None
        self.metrics = {}
        self.cond_features = {}
        self.is_cond = False
        self.inputs = list()
        self.encoders = list()
        self.feature_decoders = list()
        self.feature_losses = dict()
        self.feature_types = dict()
        self.loss_models = dict()
        self.cond_inputs = list()
        self.global_decoder = None
        self.generator = None

    def sample_z(self, args):
        mu, log_sigma = args
        eps = tf.random.normal(shape=(self.latent_dim,), mean=0.0, stddev=1.0)
        return mu + tf.exp(log_sigma / 2) * eps

    @staticmethod
    @slugify_parameters(exclude_params=("feature",))
    def _create_feature_loss_layer(feature, name):
        FeatureLossLayer(feature, name=name)

    def build_model(self):
        for name, feature in self.dataset.features.items():
            if name in self.cond_features and self.is_cond:
                self.cond_inputs.append(feature.input)

            self.inputs.append(feature.input)
            self.encoders.append(feature.encoder)

        if len(self.encoders) > 1:
            features_encoders = concatenate(self.encoders)
        else:
            features_encoders = self.encoders[0]

        self.mu, self.log_sigma = self.__build_encoder(features_encoders)
        # embed_layer = SampleLayer(gamma=2,
        #                          capacity=30,
        #                          name='sampling_layer')([self.mu, self.log_sigma])
        z = Lambda(self.sample_z)([self.mu, self.log_sigma])

        if self.is_cond:
            if len(self.cond_inputs) > 1:
                self.cond_input = concatenate(self.cond_inputs)
                logger.info(f"Conditioning on {set(self.cond_features)}")
            else:
                self.cond_input = self.cond_inputs[0]
            z_cond = concatenate([z, self.cond_input])

            gen_inp_shape = self.latent_dim + self.cond_input.shape[1]
            decoder_input = z_cond
            encoder_output = z_cond
        else:
            gen_inp_shape = self.latent_dim
            # decoder_input = Lambda(lambda x: x, name='decoder_inp')(embed_layer)
            # encoder_output = embed_layer
            decoder_input = z
            encoder_output = self.mu

        kl_loss = (
            1
            * 0.5
            * tf.reduce_sum(tf.exp(self.log_sigma) + self.mu**2 - 1.0 - self.log_sigma, 1)
        )

        generator_input = Input(shape=(gen_inp_shape,))

        self.__build_decoder(decoder_input, generator_input)

        generator_outputs = list()
        for i, (name, feature) in enumerate(self.dataset.features.items()):
            feature_decoder = feature.create_decoder(self.global_decoder)

            self._create_feature_loss_layer(feature=feature, name=name)
            feature_tensor = feature_decoder
            self.feature_losses[name] = feature.loss
            self.feature_types[name] = feature.feature_type

            self.feature_decoders.append(feature_tensor)

            generator_outputs.append(feature.create_decoder(self.generator))

        self.model = Model(self.inputs, self.feature_decoders)
        losses = list(self.feature_losses.values())
        self.model.add_loss(losses)
        self.model.add_loss(kl_loss * 0)

        self.encoder_model = Model(self.inputs, encoder_output)

        # generator
        self.generator_model = Model(generator_input, generator_outputs)

        return self.model

    def __build_encoder(self, input):
        h0 = Dense(self.intermediate_dim, name="Encoder_0")(input)
        h0 = BatchNormalization(name="First_encoder_BN")(h0)
        h0 = Activation(tf.nn.leaky_relu)(h0)
        h0 = Dropout(0.2)(h0)

        h1 = Dense(self.intermediate_dim, name="Encoder_1")(h0)
        h1 = BatchNormalization(name="Second_encoder_BN")(h1)
        h1 = Activation(tf.nn.leaky_relu)(h1)
        h1 = Dropout(0.2)(h1)

        h2 = Dense(self.intermediate_dim, name="Encoder_2")(h1)
        h2 = BatchNormalization(name="Third_encoder_BN")(h2)
        h2 = Activation(tf.nn.leaky_relu)(h2)
        h2 = Dropout(0.2)(h2)

        mu = Dense(self.latent_dim, name="mu")(h2)
        log_sigma = Dense(self.latent_dim, name="log_sigma")(h2)
        return mu, log_sigma

    def __build_decoder(self, input_z, generator_input):
        if self.is_cond:
            decoder_h0 = Dense(
                self.intermediate_dim + self.cond_input.shape[1],
                activation=LeakyReLU(),
                name="Decoder_0",
            )
        else:
            decoder_h0 = Dense(self.intermediate_dim, activation=LeakyReLU(), name="Decoder_0")
        decoder_h1 = Dense(self.intermediate_dim, activation=LeakyReLU(), name="Decoder_1")
        decoder_h2 = Dense(self.intermediate_dim, activation=LeakyReLU(), name="Decoder_2")

        h_decoded0 = Dropout(0.2)(decoder_h0(input_z))
        h_decoded1 = Dropout(0.2)(decoder_h1(h_decoded0))
        self.global_decoder = Dropout(0.2)(decoder_h2(h_decoded1))

        generator0 = decoder_h0(generator_input)
        generator1 = decoder_h1(generator0)
        self.generator = decoder_h2(generator1)

    def fit(self, data: pd.DataFrame, **kwargs) -> dict:
        transformed_data = self.dataset.transform(data)
        return self.model.fit(transformed_data, batch_size=self.batch_size, **kwargs)

    def fit_sampler(self, data: pd.DataFrame):
        log_message = "Fit sampler"
        logger.info(log_message)
        ProgressBarHandler().set_progress(message=log_message)
        transformed_data = self.dataset.transform(data)
        log_message = "Start encoding"
        logger.info(log_message)
        ProgressBarHandler().set_progress(message=log_message)
        latent_points = self.encoder_model.predict(transformed_data)

        logger.info("Creating BayesianGaussianMixture")
        self.latent_model = BayesianGaussianMixture(n_components=self.latent_components, n_init=10)
        logger.info("Fitting BayesianGaussianMixture")
        self.latent_model.fit(latent_points)
        logger.info("Finished fitting BayesianGaussianMixture")

    def predict(self, data: pd.DataFrame) -> pd.DataFrame:
        transformed_data = self.dataset.transform(data)
        prediction = self.model.predict(transformed_data, batch_size=self.batch_size)
        return self.dataset.inverse_transform(prediction)

    def sample(self, nb_samples: int) -> pd.DataFrame:
        latent_sample = self.latent_model.sample(nb_samples)[0]
        np.random.shuffle(latent_sample)

        synthetic_prediction = self.generator_model.predict(latent_sample)
        self.inverse_transformed_df = self.dataset.inverse_transform(synthetic_prediction)
        pk_uq_keys_mapping = self.dataset.primary_keys_mapping
        if pk_uq_keys_mapping:
            self.__make_pk_uq_unique(pk_uq_keys_mapping, self.dataset.dropped_columns)
        return self.inverse_transformed_df

    def less_likely_sample(
        self, nb_samples: int, temp: float = 0.05, variaty: int = 3
    ) -> pd.DataFrame:
        def pop_npoints(latent_vector, log):
            log_probs = {prob: idx for idx, prob in enumerate(log)}
            sorted_log_probs_keys = np.sort(list(log_probs.keys()))
            sorted_samples = {k: log_probs[k] for k in sorted_log_probs_keys}
            idxs = list(sorted_samples.values())[: int(nb_samples * temp / variaty)]
            return latent_vector[idxs]

        sliced_latent_sample = []
        for i in range(variaty):
            latent_sample = self.latent_model.sample(nb_samples)[0]
            log_likelihoods = self.latent_model.score_samples(latent_sample)
            sliced_latent_sample.append(pop_npoints(latent_sample, log_likelihoods))

        synthetic_prediction = self.generator_model.predict(sliced_latent_sample)
        return self.dataset.inverse_transform(synthetic_prediction)

    def __check_pk_numeric_convertability(self, column, key_type):
        if (
            key_type is str
            and column not in self.dataset.long_text_columns | self.dataset.uuid_columns
        ):
            return self.inverse_transformed_df[column].dropna().str.isnumeric().all()
        else:
            return False

    def __make_pk_uq_unique(self, pk_uq_keys_mapping, empty_columns):
        for key_name, config in pk_uq_keys_mapping.items():
            key_columns = [
                column for column in config.get("columns") if column not in empty_columns
            ]
            for column in key_columns:
                key_type = self.dataset.pk_uq_keys_types[column]
                if key_type is float or self.__check_pk_numeric_convertability(column, key_type):
                    mapped_keys = np.arange(len(self.inverse_transformed_df[column])) + 1
                    self.inverse_transformed_df[column] = mapped_keys

    def save_state(self, path: str):
        pth = Path(path)

        if self.model is not None:
            self.model.save_weights(str(pth / "vae.ckpt"))

        if self.generator_model is not None:
            self.generator_model.save_weights(str(pth / "vae_generator.ckpt"))

        if self.latent_model is not None:
            with open(str(pth / "latent_model.pkl"), "wb") as f:
                f.write(pickle.dumps(self.latent_model))

    def load_state(self, path: str):
        pth = Path(path)
        self.model.load_weights(str(pth / "vae.ckpt"))
        self.generator_model.load_weights(str(pth / "vae_generator.ckpt"))

        with open(str(pth / "latent_model.pkl"), "rb") as f:
            self.latent_model = pickle.loads(f.read())



from typing import Dict

from tensorflow.keras.layers import Layer
import tensorflow.keras.backend as K


class FeatureLossLayer(Layer):
    def __init__(self, feature, **kwargs):
        self.feature = feature
        super().__init__(**kwargs)

    def call(self, inputs, **kwargs):
        feature_input, feature_decoder = inputs
        self.add_loss(self.feature.loss, inputs=inputs)
        return feature_decoder


class SampleLayer(Layer):
    def __init__(self, gamma, capacity, **kwargs):
        super().__init__(**kwargs)
        self.gamma = gamma
        self.max_capacity = capacity

    def build(self, input_shape):
        super(SampleLayer, self).build(input_shape)
        self.built = True

    def call(self, layer_inputs, **kwargs):
        if len(layer_inputs) != 2:
            raise Exception("input layers must be a list: mean and stddev")
        if len(K.int_shape(layer_inputs[0])) != 2 or len(K.int_shape(layer_inputs[1])) != 2:
            raise Exception("input shape is not a vector [batchSize, latentSize]")

        mean = layer_inputs[0]
        log_var = layer_inputs[1]

        batch = K.shape(mean)[0]
        dim = K.int_shape(mean)[1]

        latent_loss = -0.5 * (1 + log_var - K.square(mean) - K.exp(log_var))
        latent_loss = K.sum(latent_loss, axis=1, keepdims=True)
        latent_loss = K.mean(latent_loss)
        latent_loss = self.gamma * K.abs(latent_loss - self.max_capacity)

        latent_loss = K.reshape(latent_loss, [1, 1])

        epsilon = K.random_normal(shape=(batch, dim), mean=0.0, stddev=1.0)
        layer_output = mean + K.exp(0.5 * log_var) * epsilon

        self.add_loss(losses=[latent_loss], inputs=[layer_inputs])

        return layer_output

    def compute_output_shape(self, input_shape):
        return input_shape[0]

    @property
    def get_config(self):
        config = {
            "gamma": self.gamma,
            "capacity": self.max_capacity,
        }
        base_config: Dict = super(SampleLayer, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



from syngen.ml.metrics.metrics_classes.metrics import (  # noqa: F401;
    BaseMetric,
    BivariateMetric,
    UnivariateMetric,
    JensenShannonDistance,
    Correlations,
    Clustering,
    Utility
)
from syngen.ml.metrics.accuracy_test.accuracy_test import (  # noqa: F401;
    BaseTest,
    AccuracyTest
)
from syngen.ml.metrics.sample_test.sample_test import SampleAccuracyTest  # noqa: F401
from syngen.ml.metrics.utils import encode_categorical_features  # noqa: F401
